from gufe.tokenization import (
    GufeTokenizable,
    TOKENIZABLE_REGISTRY,
    get_all_gufe_objs,
    is_gufe_obj,
    modify_dependencies,
)
from itertools import chain
from typing import Union, Dict, List
import networkx as nx


class RegistryBackup(object):
    def __init__(self, gufe_object=None, keep_changes=False):
        self.gufe_object = gufe_object
        self.keep_changes = keep_changes

    def __enter__(self):
        self.registry_backup = TOKENIZABLE_REGISTRY.copy()
        self.clear_gufe_deps()
        return self.registry_backup

    def __exit__(self, exception, value, traceback):
        if not self.keep_changes or exception is not None:
            TOKENIZABLE_REGISTRY.clear()
        TOKENIZABLE_REGISTRY.update(self.registry_backup)

    def clear_gufe_deps(self):
        if self.gufe_object is not None:
            objects = get_all_gufe_objs(self.gufe_object)

            for o in objects:
                gufe_key = o.key
                TOKENIZABLE_REGISTRY.pop(gufe_key)
        else:
            TOKENIZABLE_REGISTRY.clear()


def gufe_objects_from_shallow_dict(
    obj: Union[List, Dict, GufeTokenizable]
) -> List[GufeTokenizable]:
    """Find GufeTokenizables within a shallow dict.

    This function recursively looks through the list/dict structures encoding
    GufeTokenizables and returns list of all GufeTokenizables found
    within those structures, which may be potentially nested.

    Parameters
    ----------
    obj
        The input data structure to recursively traverse. For the initial call
        of this function, this should be the shallow dict of a GufeTokenizable.
        Input of a GufeTokenizable will immediately return a base case.

    Returns
    -------
    List[GufeTokenizable]
        All GufeTokenizables found in the shallow dict representation of a
        GufeTokenizable.

    """
    if is_gufe_obj(obj):
        return [obj]

    elif isinstance(obj, list):
        return list(
            chain.from_iterable([gufe_objects_from_shallow_dict(item) for item in obj])
        )

    elif isinstance(obj, dict):
        return list(
            chain.from_iterable(
                [gufe_objects_from_shallow_dict(item) for item in obj.values()]
            )
        )

    return []


def gufe_to_digraph(gufe_obj):
    """Recursively construct a DiGraph from a GufeTokenizable.

    The DiGraph encodes the dependency structure of the GufeTokenizable on
    other GufeTokenizables.
    """
    graph = nx.DiGraph()
    shallow_dicts = {}

    def add_edges(o):
        # if we've made a shallow dict before, we've already added this one
        # and all its dependencies; return `None` to avoid going down the tree
        # again
        sd = shallow_dicts.get(o.key)
        if sd is not None:
            return None

        # if not, then we make the shallow dict only once, add it to our index,
        # add edges to dependencies, and return it so we continue down the tree
        sd = o.to_shallow_dict()

        shallow_dicts[o.key] = sd

        # add the object node in case there aren't any connections
        graph.add_node(o)
        connections = gufe_objects_from_shallow_dict(sd)

        for c in connections:
            graph.add_edge(o, c)

        return sd

    sd = add_edges(gufe_obj)
    _ = modify_dependencies(sd, add_edges, is_gufe_obj, mode="encode")

    return graph



"""
:mod:`alchemiscale.models` --- common data models
=================================================

"""
from typing import Optional, Union
from pydantic import BaseModel, Field, validator, root_validator
from gufe.tokenization import GufeKey
from re import fullmatch


class Scope(BaseModel):
    org: Optional[str] = None
    campaign: Optional[str] = None
    project: Optional[str] = None

    def __init__(self, org=None, campaign=None, project=None):
        # we add this to allow for arg-based creation, not just keyword-based
        super().__init__(org=org, campaign=campaign, project=project)

    def __str__(self):
        triple = (
            i if i is not None else "*" for i in (self.org, self.campaign, self.project)
        )
        return "-".join(triple)

    def __lt__(self, other):
        return str(self) < str(other)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return str(self) == str(other)

    class Config:
        frozen = True

    @staticmethod
    def _validate_component(v, component):
        """
        Use regex to check that the component:
        - contains only alphanumeric or underscore characters, or is just a single asterisk
        - does not start with an underscore or number
        """

        # we require that there is a full match, so that the string is not
        # allowed to contain any other characters.
        if v is not None and not fullmatch(r"^[a-zA-Z][a-zA-Z0-9_]*|\*$", v):
            raise InvalidScopeError(
                f"'{component}' must either start with an alphabetical and contain "
                "only alphanumeric or underscore ('_') thereafter "
                "OR must be a single asterisk ('*')"
            )

        if v == "*":
            # if we're given an asterisk, cast this to `None` instead for
            # consistency
            v = None

        return v

    @validator("org")
    def valid_org(cls, v):
        return cls._validate_component(v, "org")

    @validator("campaign")
    def valid_campaign(cls, v):
        return cls._validate_component(v, "campaign")

    @validator("project")
    def valid_project(cls, v):
        return cls._validate_component(v, "project")

    @root_validator
    def check_scope_hierarchy(cls, values):
        if not _hierarchy_valid(values):
            raise InvalidScopeError(
                f"Invalid scope hierarchy: {values}, cannot specify wildcard ('*')"
                " in a scope component if a less specific scope component is not"
                " given, unless all components are wildcards (*-*-*)."
            )
        return values

    def to_tuple(self):
        return (self.org, self.campaign, self.project)

    @classmethod
    def from_str(cls, string):
        org, campaign, project = (i if i != "*" else None for i in string.split("-"))
        return cls(org=org, campaign=campaign, project=project)

    def is_superset(self, other: "Scope") -> bool:
        """Return `True` if this Scope is a superset of another.

        Check for a superset (not a proper superset) so that two equal scopes
        also return `True`.

        """
        if self.org is not None and self.org != other.org:
            return False
        if self.campaign is not None and self.campaign != other.campaign:
            return False
        if self.project is not None and self.project != other.project:
            return False
        return True

    def __repr__(self):  # pragma: no cover
        return f"<Scope('{str(self)}')>"

    def specific(self) -> bool:
        """Return `True` if this Scope has no unspecified elements."""
        return all(self.to_tuple())


class ScopedKey(BaseModel):
    """Unique identifier for GufeTokenizables in state store.

    For this object, `org`, `campaign`, and `project` cannot contain wildcards.
    In other words, the Scope of a ScopedKey must be *specific*.

    """

    gufe_key: GufeKey
    org: str
    campaign: str
    project: str

    class Config:
        frozen = True

    @validator("gufe_key")
    def cast_gufe_key(cls, v):
        return GufeKey(v)

    def __repr__(self):  # pragma: no cover
        return f"<ScopedKey('{str(self)}')>"

    def __str__(self):
        return "-".join([self.gufe_key, self.org, self.campaign, self.project])

    def __lt__(self, other):
        return str(self) < str(other)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return str(self) == str(other)

    @classmethod
    def from_str(cls, string):
        prefix, token, org, campaign, project = string.split("-")
        gufe_key = GufeKey(f"{prefix}-{token}")

        return cls(gufe_key=gufe_key, org=org, campaign=campaign, project=project)

    @property
    def scope(self):
        return Scope(org=self.org, campaign=self.campaign, project=self.project)

    @property
    def qualname(self):
        return self.gufe_key.split("-")[0]

    def to_dict(self):
        return self.dict()

    @classmethod
    def from_dict(cls, d):
        return cls(**d)


class InvalidScopeError(ValueError):
    ...


def _is_wildcard(char: Union[str, None]) -> bool:
    return char is None


def _find_wildcard(scope_list: list) -> Union[int, None]:
    """Finds the index of the first wildcard in a scope list."""
    for i, scope in enumerate(scope_list):
        if _is_wildcard(scope):
            return i
    return None


def _hierarchy_valid(scope_dict: dict[str : Union[str, None]]) -> bool:
    """Checks that the scope hierarchy is valid from a dictionary of scope components."""

    org = scope_dict.get("org")
    campaign = scope_dict.get("campaign")
    project = scope_dict.get("project")
    scope_list = [org, campaign, project]

    first_wildcard_ix = _find_wildcard(scope_list)
    if first_wildcard_ix is None:  # no wildcards, so we're good
        return True

    sublevels = scope_list[first_wildcard_ix:]
    # now check if any of the sublevels are not wildcards
    if any([not _is_wildcard(i) for i in sublevels]):
        return False
    return True



"""
:mod:`alchemiscale.base.client` --- base class for API clients
==============================================================

"""

import asyncio
import time
import random
from itertools import islice
from typing import List
import json
from urllib.parse import urljoin
from functools import wraps
import gzip

import requests
import httpx

from gufe.tokenization import GufeTokenizable, JSON_HANDLER

from ..models import Scope, ScopedKey
from ..storage.models import TaskHub, Task


def json_to_gufe(jsondata):
    return GufeTokenizable.from_dict(json.loads(jsondata, cls=JSON_HANDLER.decoder))


class AlchemiscaleBaseClientError(Exception):
    def __init__(self, *args, **kwargs):
        self.status_code = kwargs.pop("status_code", None)
        super().__init__(*args, **kwargs)


class AlchemiscaleConnectionError(Exception):
    ...


def use_session(f):
    async def _wrapper(self, *args, **kwargs):
        self._lock = asyncio.Lock()
        self._session = httpx.AsyncClient(verify=self.verify)
        try:
            return await f(self, *args, **kwargs)
        finally:
            await self._session.aclose()
            self._session = None
            self._lock = None

    return _wrapper


class AlchemiscaleBaseClient:
    """Base class for Alchemiscale API clients."""

    _exception = AlchemiscaleBaseClientError
    _retry_status_codes = [404, 502, 503, 504]

    def __init__(
        self,
        api_url: str,
        identifier: str,
        key: str,
        max_retries: int = 5,
        retry_base_seconds: float = 2.0,
        retry_max_seconds: float = 60.0,
        verify: bool = True,
    ):
        """Client class for interfacing with an alchemiscale API service.

        Parameters
        ----------
        api_url
            URL of the API to interact with.
        identifier
            Identifier for the identity used for authentication.
        key
            Credential for the identity used for authentication.
        max_retries
            Maximum number of times to retry a request. In the case the API
            service is unresponsive an exponential backoff is applied with
            retries until this number is reached. A ``self._exception`` is
            raised if retries are exhausted. If set to -1, retries will
            continue indefinitely until success.
        retry_base_seconds
            The base number of seconds to use for exponential backoff.
            Must be greater than 1.0.
        retry_max_seconds
            Maximum number of seconds to sleep between retries; avoids runaway
            exponential backoff while allowing for many retries.
        verify
            Whether to verify SSL certificate presented by the API server.

        """
        self.api_url = api_url
        self.identifier = identifier
        self.key = key
        self.max_retries = max_retries

        if retry_base_seconds <= 1.0:
            raise ValueError("'retry_base_seconds' must be greater than 1.0")

        self.retry_base_seconds = retry_base_seconds
        self.retry_max_seconds = retry_max_seconds

        self.verify = verify

        self._jwtoken = None
        self._headers = {}

        self._session = None
        self._lock = None

    def _settings(self):
        return dict(
            api_url=self.api_url,
            identifier=self.identifier,
            key=self.key,
            max_retries=self.max_retries,
            retry_base_seconds=self.retry_base_seconds,
            retry_max_seconds=self.retry_max_seconds,
            verify=self.verify,
        )

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.api_url}')"

    def _retry(f):
        """Automatically retry with exponential backoff if API service is
        unreachable or unable to service request.

        Will retry up to ``self.max_retries``, with the time between retries
        increasing by ``self.retry_base_seconds`` ** retries plus a random
        jitter scaled to ``self.retry_base_seconds``. ``self.retry_max_seconds`
        gives an upper bound to the time between retries.

        """

        @wraps(f)
        def _wrapper(self, *args, **kwargs):
            retries = 0
            while True:
                try:
                    return f(self, *args, **kwargs)
                except (self._exception, AlchemiscaleConnectionError) as e:
                    # if we are getting back HTTP errors and the status code is not
                    # one of those we want to retry on, just raise
                    if isinstance(e, self._exception):
                        if e.status_code not in self._retry_status_codes:
                            raise

                    if (self.max_retries != -1) and retries >= self.max_retries:
                        raise
                    retries += 1

                    # apply exponential backoff with random jitter
                    sleep_time = min(
                        self.retry_max_seconds
                        + self.retry_base_seconds * random.random(),
                        self.retry_base_seconds**retries
                        + self.retry_base_seconds * random.random(),
                    )
                    time.sleep(sleep_time)

        return _wrapper

    def _retry_async(f):
        """Automatically retry with exponential backoff if API service is
        unreachable or unable to service request.

        Will retry up to ``self.max_retries``, with the time between retries
        increasing by ``self.retry_base_seconds`` ** retries plus a random
        jitter scaled to ``self.retry_base_seconds``. ``self.retry_max_seconds`
        gives an upper bound to the time between retries.

        """

        async def _wrapper(self, *args, **kwargs):
            retries = 0
            while True:
                try:
                    return await f(self, *args, **kwargs)
                except (self._exception, AlchemiscaleConnectionError) as e:
                    # if we are getting back HTTP errors and the status code is not
                    # one of those we want to retry on, just raise
                    if isinstance(e, self._exception):
                        if e.status_code not in self._retry_status_codes:
                            raise

                    if (self.max_retries != -1) and retries >= self.max_retries:
                        raise
                    retries += 1

                    # apply exponential backoff with random jitter
                    sleep_time = min(
                        self.retry_max_seconds
                        + self.retry_base_seconds * random.random(),
                        self.retry_base_seconds**retries
                        + self.retry_base_seconds * random.random(),
                    )
                    time.sleep(sleep_time)

        return _wrapper

    def _get_token(self):
        data = {"username": self.identifier, "password": self.key}

        url = urljoin(self.api_url, "/token")
        try:
            resp = requests.post(url, data=data, timeout=None, verify=self.verify)
        except requests.exceptions.RequestException as e:
            raise AlchemiscaleConnectionError(*e.args)

        if not 200 <= resp.status_code < 300:
            raise self._exception(
                f"Status Code {resp.status_code} : {resp.reason}",
                status_code=resp.status_code,
            )

        self._jwtoken = resp.json()["access_token"]
        self._headers = {
            "Authorization": f"Bearer {self._jwtoken}",
            "Content-type": "application/json",
        }

    def _use_token(f):
        @wraps(f)
        def _wrapper(self, *args, **kwargs):
            # if we don't have a token at all, get one
            if self._jwtoken is None:
                self._get_token()

            # execute our function
            # if we get an unauthorized exception, it may be that our token is
            # stale; get a new one if so
            # if it's any other status code, raise it
            try:
                return f(self, *args, **kwargs)
            except self._exception as e:
                if e.status_code == 401:
                    self._get_token()
                else:
                    raise

            # if we made it here, it means we got a fresh token;
            # try the function again with the token in place
            return f(self, *args, **kwargs)

        return _wrapper

    async def _get_token_async(self):
        data = {"username": self.identifier, "password": self.key}

        url = urljoin(self.api_url, "/token")
        try:
            resp = await self._session.post(url, data=data, timeout=None)
        except httpx.RequestError as e:
            raise AlchemiscaleConnectionError(*e.args)

        if not 200 <= resp.status_code < 300:
            raise self._exception(
                f"Status Code {resp.status_code} : {resp.reason_phrase}",
                status_code=resp.status_code,
            )

        self._jwtoken = resp.json()["access_token"]
        self._headers = {
            "Authorization": f"Bearer {self._jwtoken}",
            "Content-Type": "application/json",
        }

    def _use_token_async(f):
        async def _wrapper(self, *args, **kwargs):
            # if we don't have a token at all, get one; if a coroutine already
            # has the lock, wait until it releases
            if self._jwtoken is None:
                if not self._lock.locked():
                    await self._lock.acquire()
                    try:
                        await self._get_token_async()
                    finally:
                        self._lock.release()
                else:
                    async with self._lock:
                        pass

            # execute our function
            # if we get an unauthorized exception, it may be that our token is
            # stale; get a new one if so
            # if it's any other status code, raise it
            try:
                return await f(self, *args, **kwargs)
            except self._exception as e:
                if e.status_code == 401:
                    if not self._lock.locked():
                        await self._lock.acquire()
                        try:
                            await self._get_token_async()
                        finally:
                            self._lock.release()
                    else:
                        async with self._lock:
                            pass
                else:
                    raise

            # if we made it here, it means we got a fresh token;
            # try the function again with the token in place
            return await f(self, *args, **kwargs)

        return _wrapper

    @_retry
    @_use_token
    def _query_resource(self, resource, params=None):
        if params is None:
            params = {}

        url = urljoin(self.api_url, resource)
        try:
            resp = requests.get(
                url, params=params, headers=self._headers, verify=self.verify
            )
        except requests.exceptions.RequestException as e:
            raise AlchemiscaleConnectionError(*e.args)

        if not 200 <= resp.status_code < 300:
            raise self._exception(
                f"Status Code {resp.status_code} : {resp.reason}",
                status_code=resp.status_code,
            )

        if params.get("return_gufe"):
            return {
                ScopedKey.from_str(k): json_to_gufe(v) for k, v in resp.json().items()
            }
        else:
            return [ScopedKey.from_str(i) for i in resp.json()]

    @_retry
    @_use_token
    def _get_resource(self, resource, params=None, compress=False):
        if params is None:
            params = {}

        if compress:
            headers = self._headers | {"Accept-Encoding": "gzip"}
        else:
            headers = self._headers | {"Accept-Encoding": ""}

        url = urljoin(self.api_url, resource)
        try:
            resp = requests.get(url, params=params, headers=headers, verify=self.verify)
        except requests.exceptions.RequestException as e:
            raise AlchemiscaleConnectionError(*e.args)

        if not 200 <= resp.status_code < 300:
            try:
                detail = resp.json()["detail"]
            except:
                detail = resp.text
            raise self._exception(
                f"Status Code {resp.status_code} : {resp.reason} : {detail}",
                status_code=resp.status_code,
            )

        content = json.loads(resp.text, cls=JSON_HANDLER.decoder)
        return content

    @_retry_async
    @_use_token_async
    async def _get_resource_async(self, resource, params=None, compress=False):
        if params is None:
            params = {}
        else:
            # drop params that are None
            params = {k: v for k, v in params.items() if v is not None}

        if compress:
            headers = self._headers | {"Accept-Encoding": "gzip"}
        else:
            headers = self._headers | {"Accept-Encoding": ""}

        url = urljoin(self.api_url, resource)
        try:
            resp = await self._session.get(
                url, params=params, headers=headers, timeout=None
            )
        except httpx.RequestError as e:
            raise AlchemiscaleConnectionError(*e.args)

        if not 200 <= resp.status_code < 300:
            try:
                detail = resp.json()["detail"]
            except:
                detail = resp.text
            raise self._exception(
                f"Status Code {resp.status_code} : {resp.reason_phrase} : {detail}",
                status_code=resp.status_code,
            )
        content = json.loads(resp.text, cls=JSON_HANDLER.decoder)
        return content

    def _post_resource(self, resource, data, compress=False):
        url = urljoin(self.api_url, resource)

        if compress is True:
            compress = 5
        elif compress is False:
            compress = 0

        if 0 < compress <= 9:
            headers = {"Content-Encoding": "gzip"}
            data_ = gzip.compress(
                json.dumps(data, cls=JSON_HANDLER.encoder).encode("utf-8"),
                mtime=0,
                compresslevel=compress,
            )
        elif compress == 0:
            headers = {}
            data_ = json.dumps(data, cls=JSON_HANDLER.encoder)
        else:
            ValueError(
                "`compress` must be a bool or an integer between 0 and 9, inclusive"
            )

        return self._post(url, headers, data_)

    @_retry
    @_use_token
    def _post(self, url, headers, data):
        headers = dict(self._headers) | headers
        try:
            resp = requests.post(url, data=data, headers=headers, verify=self.verify)
        except requests.exceptions.RequestException as e:
            raise AlchemiscaleConnectionError(*e.args)

        if not 200 <= resp.status_code < 300:
            try:
                detail = resp.json()["detail"]
            except:
                detail = resp.text
            raise self._exception(
                f"Status Code {resp.status_code} : {resp.reason} : {detail}",
                status_code=resp.status_code,
            )

        return resp.json()

    @_retry_async
    @_use_token_async
    async def _post_resource_async(self, resource, data):
        url = urljoin(self.api_url, resource)
        jsondata = json.dumps(data, cls=JSON_HANDLER.encoder)
        try:
            resp = await self._session.post(
                url, data=jsondata, headers=self._headers, timeout=None
            )
        except httpx.RequestError as e:
            raise AlchemiscaleConnectionError(*e.args)

        if not 200 <= resp.status_code < 300:
            try:
                detail = resp.json()["detail"]
            except:
                detail = resp.text
            raise self._exception(
                f"Status Code {resp.status_code} : {resp.reason_phrase} : {detail}",
                status_code=resp.status_code,
            )

        return resp.json()

    @staticmethod
    def _batched(iterable, n):
        # batched('ABCDEFG', 3) --> ABC DEF G
        if n < 1:
            raise ValueError("n must be at least one")
        it = iter(iterable)
        while batch := tuple(islice(it, n)):
            yield batch

    @staticmethod
    def _rich_waiting_columns():
        from rich.progress import SpinnerColumn, TimeElapsedColumn, TextColumn

        return [
            SpinnerColumn("dots2", style="#ff073a"),
            SpinnerColumn("dots2", style="#1793d0"),
            TextColumn("[progress.description]{task.description}"),
            TimeElapsedColumn(),
        ]

    @staticmethod
    def _rich_progress_columns():
        from rich.progress import (
            Progress,
            SpinnerColumn,
            MofNCompleteColumn,
            TextColumn,
            BarColumn,
            TimeRemainingColumn,
        )

        return [
            SpinnerColumn("dots2", style="#ff073a"),
            SpinnerColumn("dots2", style="#1793d0"),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            MofNCompleteColumn(),
            TimeRemainingColumn(),
        ]

    @_retry
    def get_info(self):
        return self._get_resource("/info")

    @_retry
    @_use_token
    def _api_check(self):
        # Check if the API is up and running and can reach services
        self._get_resource("/check")



"""
:mod:`alchemiscale.settings` --- settings
=========================================

"""

from functools import lru_cache
from typing import Optional

from pydantic import BaseSettings


class FrozenSettings(BaseSettings):
    class Config:
        frozen = True


class Neo4jStoreSettings(FrozenSettings):
    """Automatically populates settings from environment variables where they
    match; case-insensitive.

    """

    NEO4J_URL: str
    NEO4J_DBNAME: str = "neo4j"
    NEO4J_USER: str
    NEO4J_PASS: str


class S3ObjectStoreSettings(FrozenSettings):
    """Automatically populates settings from environment variables where they
    match; case-insensitive.

    If deploying APIs that use the S3ObjectStore on an EC2 host or other
    role-based resource (e.g. an ECS container), then don't set these.
    Instead rely on the IAM role of that resource for granting access to S3.

    """

    AWS_ACCESS_KEY_ID: Optional[str] = None
    AWS_SECRET_ACCESS_KEY: Optional[str] = None
    AWS_SESSION_TOKEN: Optional[str] = None
    AWS_S3_BUCKET: str
    AWS_S3_PREFIX: str
    AWS_DEFAULT_REGION: str


class JWTSettings(FrozenSettings):
    """Automatically populates settings from environment variables where they
    match; case-insensitive.

    """

    JWT_SECRET_KEY: str
    JWT_EXPIRE_SECONDS: int = 1800
    JWT_ALGORITHM: str = "HS256"


class BaseAPISettings(Neo4jStoreSettings, S3ObjectStoreSettings, JWTSettings):
    ...


class APISettings(BaseAPISettings):
    """Automatically populates settings from environment variables where they
    match; case-insensitive.

    """

    ALCHEMISCALE_API_HOST: str = "127.0.0.1"
    ALCHEMISCALE_API_PORT: int = 80
    ALCHEMISCALE_API_LOGLEVEL: str = "info"


class ComputeAPISettings(BaseAPISettings):
    """Automatically populates settings from environment variables where they
    match; case-insensitive.

    """

    ALCHEMISCALE_COMPUTE_API_HOST: str = "127.0.0.1"
    ALCHEMISCALE_COMPUTE_API_PORT: int = 80
    ALCHEMISCALE_COMPUTE_API_LOGLEVEL: str = "info"
    ALCHEMISCALE_COMPUTE_API_REGISTRATION_EXPIRE_SECONDS: int = 1800


@lru_cache()
def get_neo4jstore_settings():
    return Neo4jStoreSettings()


@lru_cache()
def get_s3objectstore_settings():
    return S3ObjectStoreSettings()


@lru_cache()
def get_jwt_settings():
    return JWTSettings()


@lru_cache()
def get_base_api_settings():
    return BaseAPISettings()


@lru_cache()
def get_api_settings():
    return APISettings()


@lru_cache()
def get_compute_api_settings():
    return ComputeAPISettings()



"""
:mod:`alchemiscale.compute.api` --- compute API components
==========================================================

"""

from typing import Any, Dict, List, Optional
import os
import json
from datetime import datetime, timedelta

from fastapi import FastAPI, APIRouter, Body, Depends, HTTPException, status
from fastapi.middleware.gzip import GZipMiddleware
from gufe.tokenization import GufeTokenizable, JSON_HANDLER

from ..base.api import (
    QueryGUFEHandler,
    scope_params,
    get_token_data_depends,
    get_n4js_depends,
    get_s3os_depends,
    base_router,
    get_cred_entity,
    validate_scopes,
    validate_scopes_query,
    _check_store_connectivity,
    gufe_to_json,
    GzipRoute,
)
from ..settings import (
    get_base_api_settings,
    get_compute_api_settings,
    ComputeAPISettings,
)
from ..storage.statestore import Neo4jStore, get_n4js
from ..storage.objectstore import S3ObjectStore
from ..storage.models import (
    ProtocolDAGResultRef,
    ComputeServiceID,
    ComputeServiceRegistration,
    TaskStatusEnum,
)
from ..models import Scope, ScopedKey
from ..security.auth import get_token_data, oauth2_scheme
from ..security.models import (
    Token,
    TokenData,
    CredentialedComputeIdentity,
)


app = FastAPI(title="AlchemiscaleComputeAPI")
app.dependency_overrides[get_base_api_settings] = get_compute_api_settings
app.include_router(base_router)
app.add_middleware(GZipMiddleware, minimum_size=1000, compresslevel=5)


def get_cred_compute():
    return CredentialedComputeIdentity


app.dependency_overrides[get_cred_entity] = get_cred_compute


router = APIRouter(
    dependencies=[Depends(get_token_data_depends)],
)
router.route_class = GzipRoute


@app.get("/ping")
def ping():
    return {"api": "AlchemiscaleComputeAPI"}


@router.get("/info")
def info():
    return {"message": "nothing yet"}


@router.get("/check")
def check(
    n4js: Neo4jStore = Depends(get_n4js_depends),
    s3os: S3ObjectStore = Depends(get_s3os_depends),
):
    # check connectivity of storage components
    # if no exception raised, all good
    _check_store_connectivity(n4js, s3os)


@router.get("/identities/{identity_identifier}/scopes")
def list_scopes(
    *,
    identity_identifier,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[str]:
    scopes = n4js.list_scopes(identity_identifier, CredentialedComputeIdentity)
    return [str(scope) for scope in scopes]


@router.post("/computeservice/{compute_service_id}/register")
def register_computeservice(
    compute_service_id,
    n4js: Neo4jStore = Depends(get_n4js_depends),
):
    now = datetime.utcnow()
    csreg = ComputeServiceRegistration(
        identifier=compute_service_id, registered=now, heartbeat=now
    )

    compute_service_id_ = n4js.register_computeservice(csreg)

    return compute_service_id_


@router.post("/computeservice/{compute_service_id}/deregister")
def deregister_computeservice(
    compute_service_id,
    n4js: Neo4jStore = Depends(get_n4js_depends),
):
    compute_service_id_ = n4js.deregister_computeservice(
        ComputeServiceID(compute_service_id)
    )

    return compute_service_id_


@router.post("/computeservice/{compute_service_id}/heartbeat")
def heartbeat_computeservice(
    compute_service_id,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    settings: ComputeAPISettings = Depends(get_base_api_settings),
):
    now = datetime.utcnow()

    # expire any stale registrations, along with their claims
    expire_delta = timedelta(
        seconds=settings.ALCHEMISCALE_COMPUTE_API_REGISTRATION_EXPIRE_SECONDS
    )
    expire_time = now - expire_delta
    n4js.expire_registrations(expire_time)

    compute_service_id_ = n4js.heartbeat_computeservice(compute_service_id, now)

    return compute_service_id_


@router.get("/taskhubs")
def query_taskhubs(
    *,
    return_gufe: bool = False,
    scope: Scope = Depends(scope_params),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    # intersect query scopes with accessible scopes in the token
    query_scopes = validate_scopes_query(scope, token)
    taskhubs_handler = QueryGUFEHandler(return_gufe)

    # query each scope
    # loop might be more removable in the future with a Union like operator on scopes
    for single_query_scope in query_scopes:
        # add new task hubs
        taskhubs_handler.update_results(
            n4js.query_taskhubs(
                scope=single_query_scope, return_gufe=taskhubs_handler.return_gufe
            )
        )

    return taskhubs_handler.format_return()


@router.post("/taskhubs/{taskhub_scoped_key}/claim")
def claim_taskhub_tasks(
    taskhub_scoped_key,
    *,
    compute_service_id: str = Body(),
    count: int = Body(),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(taskhub_scoped_key)
    validate_scopes(sk.scope, token)

    tasks = n4js.claim_taskhub_tasks(
        taskhub=taskhub_scoped_key,
        compute_service_id=ComputeServiceID(compute_service_id),
        count=count,
    )

    return [str(t) if t is not None else None for t in tasks]


@router.get("/tasks/{task_scoped_key}/transformation")
def get_task_transformation(
    task_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    s3os: S3ObjectStore = Depends(get_s3os_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(sk.scope, token)

    transformation_sk, protocoldagresultref_sk = n4js.get_task_transformation(
        task=task_scoped_key, return_gufe=False
    )

    transformation = n4js.get_gufe(transformation_sk)

    if protocoldagresultref_sk:
        protocoldagresultref = n4js.get_gufe(protocoldagresultref_sk)
        pdr_sk = ScopedKey(gufe_key=protocoldagresultref.obj_key, **sk.scope.dict())

        # we keep this as a string to avoid useless deserialization/reserialization here
        try:
            pdr: str = s3os.pull_protocoldagresult(
                pdr_sk, transformation_sk, return_as="json", ok=True
            )
        except:
            # if we fail to get the object with the above, fall back to
            # location-based retrieval
            pdr: str = s3os.pull_protocoldagresult(
                location=protocoldagresultref.location,
                return_as="json",
                ok=True,
            )

    else:
        pdr = None

    return (gufe_to_json(transformation), pdr)


# TODO: support compression performed client-side
@router.post("/tasks/{task_scoped_key}/results", response_model=ScopedKey)
def set_task_result(
    task_scoped_key,
    *,
    protocoldagresult: str = Body(embed=True),
    compute_service_id: Optional[str] = Body(embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    s3os: S3ObjectStore = Depends(get_s3os_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    task_sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(task_sk.scope, token)

    pdr = json.loads(protocoldagresult, cls=JSON_HANDLER.decoder)
    pdr = GufeTokenizable.from_dict(pdr)

    tf_sk, _ = n4js.get_task_transformation(
        task=task_scoped_key,
        return_gufe=False,
    )

    # push the ProtocolDAGResult to the object store
    protocoldagresultref: ProtocolDAGResultRef = s3os.push_protocoldagresult(
        pdr, transformation=tf_sk, creator=compute_service_id
    )

    # push the reference to the state store
    result_sk: ScopedKey = n4js.set_task_result(
        task=task_sk, protocoldagresultref=protocoldagresultref
    )

    # if success, set task complete, remove from all hubs
    # otherwise, set as errored, leave in hubs
    if protocoldagresultref.ok:
        n4js.set_task_complete(tasks=[task_sk])
    else:
        n4js.set_task_error(tasks=[task_sk])

    return result_sk


### add router

app.include_router(router)



"""
:mod:`alchemiscale.strategies.base` --- alchemical strategy base class
======================================================================

"""

from gufe.tokenization import GufeTokenizable


class Strategy(GufeTokenizable):
    ...

    def _to_dict(self):
        ...

    def _from_dict(self):
        ...

    @property
    def _defaults(self):
        ...



"""
:mod:`alchemiscale.interface.client` --- client for interacting with user-facing API
====================================================================================

"""

import asyncio
from typing import Union, List, Dict, Optional, Tuple, Any
import json
from itertools import chain
from collections import Counter
from functools import lru_cache

import httpx
from async_lru import alru_cache
import networkx as nx
from gufe import AlchemicalNetwork, Transformation, ChemicalSystem
from gufe.tokenization import GufeTokenizable, JSON_HANDLER, GufeKey
from gufe.protocols import ProtocolResult, ProtocolDAGResult


from ..base.client import (
    AlchemiscaleBaseClient,
    AlchemiscaleBaseClientError,
    json_to_gufe,
    use_session,
)
from ..models import Scope, ScopedKey
from ..storage.models import Task, ProtocolDAGResultRef, TaskStatusEnum
from ..strategies import Strategy
from ..security.models import CredentialedUserIdentity
from ..validators import validate_network_nonself
from ..keyedchain import KeyedChain


class AlchemiscaleClientError(AlchemiscaleBaseClientError):
    ...


def _get_transformation_results(client_settings, tf_sk, ok: bool, kwargs):
    client = AlchemiscaleClient(**client_settings)
    if ok:
        return tf_sk, client.get_transformation_results(tf_sk, **kwargs)
    else:
        return tf_sk, client.get_transformation_failures(tf_sk, **kwargs)


class AlchemiscaleClient(AlchemiscaleBaseClient):
    """Client for user interaction with API service."""

    _exception = AlchemiscaleClientError

    def get_scopes(self) -> List[Scope]:
        scopes = self._get_resource(
            f"/identities/{self.identifier}/scopes",
        )
        return sorted([Scope.from_str(s) for s in scopes])

    def list_scopes(self) -> List[Scope]:
        return self.get_scopes()

    ### inputs

    def get_scoped_key(self, obj: GufeTokenizable, scope: Scope) -> ScopedKey:
        """Given any gufe object and a fully-specified Scope, return corresponding ScopedKey.

        This method does not check that this ScopedKey is represented in the database.
        It is only a convenience for properly constructing a ScopedKey from a
        gufe object and a Scope.

        """
        if scope.specific():
            return ScopedKey(gufe_key=obj.key, **scope.dict())
        else:
            raise ValueError(
                "Scope for a ScopedKey must be specific; it cannot contain wildcards."
            )

    def check_exists(self, scoped_key: ScopedKey) -> bool:
        """Returns ``True`` if the given ScopedKey represents an object in the database."""
        return self._get_resource(f"/exists/{scoped_key}")

    def create_network(
        self,
        network: AlchemicalNetwork,
        scope: Scope,
        compress: Union[bool, int] = True,
        visualize: bool = True,
    ) -> ScopedKey:
        """Submit an AlchemicalNetwork to a specific Scope.

        Parameters
        ----------
        network
            The AlchemicalNetwork to submit.
        scope
            The Scope in which to submit the AlchemicalNetwork.
            This must be a *specific* Scope; it must not contain wildcards.
        compress
            If ``True``, compress the AlchemicalNetwork client-side before
            shipping to the API service. This can reduce submission time
            depending on the bandwidth of your connection to the API service.
            Set to ``False`` to submit without compressing. This is a
            performance optimization; it has no bearing on the result of this
            method call.

            Use an integer between 0 and 9 for finer control over
            the degree of compression; 0 means no compression, 9 means max
            compression. ``True`` is synonymous with level 5 compression.
        visualize
            If ``True``, show submission progress indicator.

        Returns
        -------
        ScopedKey
            The ScopedKey of the AlchemicalNetwork.

        """
        if not scope.specific():
            raise ValueError(
                f"`scope` '{scope}' contains wildcards ('*'); `scope` must be *specific*"
            )

        validate_network_nonself(network)

        sk = self.get_scoped_key(network, scope)

        def post():
            keyed_chain = KeyedChain.gufe_to_keyed_chain_rep(network)
            data = dict(network=keyed_chain, scope=scope.dict())
            return self._post_resource("/networks", data, compress=compress)

        if visualize:
            from rich.progress import Progress

            with Progress(*self._rich_waiting_columns(), transient=False) as progress:
                task = progress.add_task(
                    f"Submitting [bold]'{sk}'[/bold]...", total=None
                )

                scoped_key = post()
                progress.start_task(task)
                progress.update(task, total=1, completed=1)
        else:
            scoped_key = post()

        return ScopedKey.from_dict(scoped_key)

    def query_networks(
        self,
        name: Optional[str] = None,
        scope: Optional[Scope] = None,
    ) -> List[ScopedKey]:
        """Query for AlchemicalNetworks, optionally by name or Scope.

        Calling this method with no query arguments will return ScopedKeys for
        all AlchemicalNetworks that are within the Scopes this user has access
        to.

        """
        if scope is None:
            scope = Scope()

        params = dict(name=name, **scope.dict())

        return self._query_resource("/networks", params=params)

    def query_transformations(
        self,
        name: Optional[str] = None,
        scope: Optional[Scope] = None,
    ) -> List[ScopedKey]:
        """Query for Transformations, optionally by name or Scope.

        Calling this method with no query arguments will return ScopedKeys for
        all Transformations that are within the Scopes this user has access to.

        """
        if scope is None:
            scope = Scope()

        params = dict(name=name, **scope.dict())

        return self._query_resource("/transformations", params=params)

    def query_chemicalsystems(
        self,
        name: Optional[str] = None,
        scope: Optional[Scope] = None,
    ) -> List[ScopedKey]:
        """Query for ChemicalSystems, optionally by name or Scope.

        Calling this method with no query arguments will return ScopedKeys for
        all ChemicalSystems that are within the Scopes this user has access to.

        """
        if scope is None:
            scope = Scope()

        params = dict(name=name, **scope.dict())

        return self._query_resource("/chemicalsystems", params=params)

    def get_network_transformations(self, network: ScopedKey) -> List[ScopedKey]:
        """List ScopedKeys for Transformations associated with the given AlchemicalNetwork."""
        return self._query_resource(f"/networks/{network}/transformations")

    def get_network_weight(self, network: ScopedKey) -> float:
        """Get the weight of the TaskHub associated with the given AlchemicalNetwork.

        Compute services perform a weighted selection of the AlchemicalNetworks
        visible to them before claiming Tasks actioned on those networks.
        Networks with higher weight are more likely to be selected than those
        with lower weight, and so will generally get more compute attention
        over time.

        A weight of ``0`` means the AlchemicalNetwork will not receive any
        compute for its actioned Tasks.

        """
        return self._get_resource(f"/networks/{network}/weight")

    def set_network_weight(self, network: ScopedKey, weight: float) -> None:
        """Set the weight of the TaskHub associated with the given AlchemicalNetwork.

        Compute services perform a weighted selection of the AlchemicalNetworks
        visible to them before claiming Tasks actioned on those networks.
        Networks with higher weight are more likely to be selected than those
        with lower weight, and so will generally get more compute attention
        over time.

        A weight of ``0`` means the AlchemicalNetwork will not receive any
        compute for its actioned Tasks.

        Parameters
        ----------
        network
            The ScopedKey of the AlchemicalNetwork to set the weight for.
        weight
            The weight to set for the network. This must be between 0 and 1
            (inclusive). Setting the value to 0 will effectively disable
            compute on this network without cancelling its actioned Tasks.

        """
        self._post_resource(f"/networks/{network}/weight", weight)

    def get_transformation_networks(self, transformation: ScopedKey) -> List[ScopedKey]:
        """List ScopedKeys for AlchemicalNetworks associated with the given Transformation."""
        return self._query_resource(f"/transformations/{transformation}/networks")

    def get_network_chemicalsystems(self, network: ScopedKey) -> List[ScopedKey]:
        """List ScopedKeys for the ChemicalSystems associated with the given AlchemicalNetwork."""
        return self._query_resource(f"/networks/{network}/chemicalsystems")

    def get_chemicalsystem_networks(self, chemicalsystem: ScopedKey) -> List[ScopedKey]:
        """List ScopedKeys for the AlchemicalNetworks associated with the given ChemicalSystem."""
        return self._query_resource(f"/chemicalsystems/{chemicalsystem}/networks")

    def get_transformation_chemicalsystems(
        self, transformation: ScopedKey
    ) -> List[ScopedKey]:
        """List ScopedKeys for the ChemicalSystems associated with the given Transformation."""
        return self._query_resource(
            f"/transformations/{transformation}/chemicalsystems"
        )

    def get_chemicalsystem_transformations(
        self, chemicalsystem: ScopedKey
    ) -> List[ScopedKey]:
        """List ScopedKeys for the Transformations associated with the given ChemicalSystem."""
        return self._query_resource(
            f"/chemicalsystems/{chemicalsystem}/transformations"
        )

    @lru_cache(maxsize=100)
    def get_network(
        self,
        network: Union[ScopedKey, str],
        compress: bool = True,
        visualize: bool = True,
    ) -> AlchemicalNetwork:
        """Retrieve an AlchemicalNetwork given its ScopedKey.

        Parameters
        ----------
        network
            The ScopedKey of the AlchemicalNetwork to retrieve.
        compress
            If ``True``, compress the AlchemicalNetwork server-side before
            shipping it to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicator.

        Returns
        -------
        AlchemicalNetwork
            The retrieved AlchemicalNetwork.

        """

        def _get_network():
            content = self._get_resource(f"/networks/{network}", compress=compress)
            return KeyedChain(content).to_gufe()

        if visualize:
            from rich.progress import (
                Progress,
                SpinnerColumn,
                TimeElapsedColumn,
                TextColumn,
            )

            with Progress(*self._rich_waiting_columns(), transient=False) as progress:
                task = progress.add_task(
                    f"Retrieving [bold]'{network}'[/bold]...", total=None
                )

                an = _get_network()

                progress.start_task(task)
                progress.update(task, total=1, completed=1)
        else:
            an = _get_network()
        return an

    @lru_cache(maxsize=10000)
    def get_transformation(
        self,
        transformation: Union[ScopedKey, str],
        compress: bool = True,
        visualize: bool = True,
    ) -> Transformation:
        """Retrieve a Transformation given its ScopedKey.

        Parameters
        ----------
        transformation
            The ScopedKey of the Transformation to retrieve.
        compress
            If ``True``, compress the Transformation server-side before
            shipping it to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicator.

        Returns
        -------
        Transformation
            The retrieved Transformation.

        """

        def _get_transformation():
            content = self._get_resource(
                f"/transformations/{transformation}", compress=compress
            )
            return KeyedChain(content).to_gufe()

        if visualize:
            from rich.progress import Progress, SpinnerColumn, TimeElapsedColumn

            with Progress(*self._rich_waiting_columns(), transient=False) as progress:
                task = progress.add_task(
                    f"Retrieving [bold]'{transformation}'[/bold]...", total=None
                )

                tf = _get_transformation()
                progress.start_task(task)
                progress.update(task, total=1, completed=1)
        else:
            tf = _get_transformation()

        return tf

    @lru_cache(maxsize=1000)
    def get_chemicalsystem(
        self,
        chemicalsystem: Union[ScopedKey, str],
        compress: bool = True,
        visualize: bool = True,
    ) -> ChemicalSystem:
        """Retrieve a ChemicalSystem given its ScopedKey.

        Parameters
        ----------
        chemicalsystem
            The ScopedKey of the ChemicalSystem to retrieve.
        compress
            If ``True``, compress the ChemicalSystem server-side before
            shipping it to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicator.

        Returns
        -------
        ChemicalSystem
            The retrieved ChemicalSystem.

        """

        def _get_chemicalsystem():
            content = self._get_resource(
                f"/chemicalsystems/{chemicalsystem}", compress=compress
            )
            return KeyedChain(content).to_gufe()

        if visualize:
            from rich.progress import Progress

            with Progress(*self._rich_waiting_columns(), transient=False) as progress:
                task = progress.add_task(
                    f"Retrieving [bold]'{chemicalsystem}'[/bold]...", total=None
                )

                cs = _get_chemicalsystem()

                progress.start_task(task)
                progress.update(task, total=1, completed=1)
        else:
            cs = _get_chemicalsystem()

        return cs

    ### compute

    def set_strategy(self, network: ScopedKey, strategy: Strategy):
        """Set the Strategy for evaluating the given AlchemicalNetwork.

        The Strategy will be applied to create and action tasks for the
        Transformations in the AlchemicalNetwork without user interaction.

        """
        raise NotImplementedError

    def create_tasks(
        self,
        transformation: ScopedKey,
        extends: Optional[ScopedKey] = None,
        count: int = 1,
    ) -> List[ScopedKey]:
        """Create Tasks for the given Transformation.

        Parameters
        ----------
        transformation
            The ScopedKey of the Transformation to create Tasks for.
        extends
            The ScopedKey for the Task to use as a starting point for the Tasks created.
        count
            The number of new Tasks to create.

        Returns
        -------
        List[ScopedKey]
            A list giving the ScopedKeys of the new Tasks created.

        """
        if extends:
            extends = extends.dict()

        data = dict(extends=extends, count=count)
        task_sks = self._post_resource(f"/transformations/{transformation}/tasks", data)
        return [ScopedKey.from_str(i) for i in task_sks]

    def query_tasks(
        self,
        scope: Optional[Scope] = None,
        status: Optional[str] = None,
    ) -> List[ScopedKey]:
        """Query for Tasks, optionally by status or Scope.

        Calling this method with no query arguments will return ScopedKeys for
        all Tasks that are within the Scopes this user has access to.

        """
        if scope is None:
            scope = Scope()

        params = dict(status=status, **scope.dict())

        return self._query_resource("/tasks", params=params)

    def get_network_tasks(self, network: ScopedKey, status: Optional[str] = None):
        """List ScopedKeys for all Tasks associated with the given AlchemicalNetwork."""
        params = {"status": status}
        return self._query_resource(f"/networks/{network}/tasks", params=params)

    def get_task_networks(self, task: ScopedKey):
        """List ScopedKeys for all AlchemicalNetworks associated with the given Task."""
        return self._query_resource(f"/tasks/{task}/networks")

    def get_transformation_tasks(
        self,
        transformation: ScopedKey,
        extends: Optional[ScopedKey] = None,
        return_as: str = "list",
        status: Optional[str] = None,
    ) -> Union[List[ScopedKey], nx.DiGraph]:
        """Return the Tasks associated with the given Transformation.

        Parameters
        ----------
        transformation
            The ScopedKey of the Transformation to get Tasks for.
        extends
            If given, only return Tasks that extend from the given Task's ScopedKey.
            This will also give any Tasks that extend from those Tasks, recursively.
            Using this keyword argument amounts to choosing the tree of Tasks that
            extend from the given Task.
        return_as : ['list', 'graph']
            If 'list', Tasks will be returned in no particular order.
            If 'graph', Tasks will be returned in a `networkx.DiGraph`, with a
            directed edge pointing from a given Task to the Task it extends.

        """
        if extends:
            extends = str(extends)

        params = dict(extends=extends, return_as=return_as, status=status)
        task_sks = self._get_resource(
            f"/transformations/{transformation}/tasks", params
        )

        if return_as == "list":
            return [ScopedKey.from_str(i) for i in task_sks]
        elif return_as == "graph":
            g = nx.DiGraph()
            for task, extends in task_sks.items():
                g.add_node(ScopedKey.from_str(task))
                if extends is not None:
                    g.add_edge(ScopedKey.from_str(task), ScopedKey.from_str(extends))

            return g
        else:
            raise ValueError(f"`return_as` takes 'list' or 'graph', not '{return_as}'")

    def get_task_transformation(self, task: ScopedKey) -> ScopedKey:
        """Get the Transformation associated with the given Task."""
        transformation = self._get_resource(f"/tasks/{task}/transformation")
        return ScopedKey.from_str(transformation)

    def _visualize_status(self, status_counts, status_object):
        from rich import print as rprint
        from rich.table import Table

        title = f"{status_object}"
        table = Table(title=title, title_justify="left", expand=True)
        # table = Table()

        table.add_column("status", justify="left", no_wrap=True)
        table.add_column("count", justify="right")

        table.add_row("complete", f"{status_counts.get('complete', 0)}", style="green")
        table.add_row("running", f"{status_counts.get('running', 0)}", style="orange3")
        table.add_row("waiting", f"{status_counts.get('waiting', 0)}", style="#1793d0")
        table.add_row("error", f"{status_counts.get('error', 0)}", style="#ff073a")
        table.add_row("invalid", f"{status_counts.get('invalid', 0)}", style="magenta1")
        table.add_row("deleted", f"{status_counts.get('deleted', 0)}", style="purple")

        rprint(table)

    def get_scope_status(
        self,
        scope: Optional[Scope] = None,
        visualize: Optional[bool] = True,
    ) -> Dict[str, int]:
        """Return status counts for all Tasks within the given Scope.

        Parameters
        ----------
        scope
            Scope to use for querying status. Non-specific Scopes are allowed,
            and will give back counts for all Tasks within that this user has
            Scope access to. Defaults to all Scopes.
        visualize
            If ``True``, print a table of status counts.

        Returns
        -------
        status_counts
            Dict giving statuses as keys, Task counts as values.
        """
        if scope is None:
            scope = Scope()

        status_counts = self._get_resource(f"/scopes/{scope}/status")

        if visualize:
            self._visualize_status(status_counts, scope)

        return status_counts

    def get_network_status(
        self,
        network: ScopedKey,
        visualize: Optional[bool] = True,
    ) -> Dict[str, int]:
        """Return status counts for all Tasks associated with the given AlchemicalNetwork.

        Parameters
        ----------
        network
            ScopedKey for the Alchemicalnetwork to obtain status counts for.
        visualize
            If ``True``, print a table of status counts.

        Returns
        -------
        status_counts
            Dict giving statuses as keys, Task counts as values.
        """
        status_counts = self._get_resource(f"/networks/{network}/status")

        if visualize:
            self._visualize_status(status_counts, network)

        return status_counts

    def get_transformation_status(
        self,
        transformation: ScopedKey,
        visualize: Optional[bool] = True,
    ) -> Dict[str, int]:
        """Return status counts for all Tasks associated with the given
        Transformation.

        Parameters
        ----------
        transformation
            ScopedKey for the Transformation to obtain status counts for.
        visualize
            If ``True``, print a table of status counts.

        Returns
        -------
        status_counts
            Dict giving statuses as keys, Task counts as values.
        """
        status_counts = self._get_resource(f"/transformations/{transformation}/status")

        if visualize:
            self._visualize_status(status_counts, transformation)

        return status_counts

    def get_network_actioned_tasks(
        self,
        network: ScopedKey,
        task_weights: bool = False,
    ) -> Union[Dict[ScopedKey, float], List[ScopedKey]]:
        """Return all actioned Tasks for a given AlchemicalNetwork.

        Parameters
        ----------
        network
            The ScopedKey for the AlchemicalNetwork to get actioned Tasks for.
        task_weights
            If ``True``, return a dict with Task ScopedKeys as keys, Task
            weights on the AlchemicalNetwork as values.

        Returns
        -------
        tasks
            A list of Task ScopedKeys actioned on the given AlchemicalNetwork.
            If ``task_weights`` is ``True``, a dict is returned with Task
            ScopedKeys as keys, Task weights as values.
        """
        data = dict(task_weights=task_weights)
        tasks = self._post_resource(f"/networks/{network}/tasks/actioned", data)

        if task_weights:
            return {ScopedKey.from_str(t): w for t, w in tasks.items()}

        return [ScopedKey.from_str(t) for t in tasks]

    def get_task_actioned_networks(
        self, task: ScopedKey, task_weights: bool = False
    ) -> Union[Dict[ScopedKey, float], List[ScopedKey]]:
        """Return all AlchemicalNetworks the given Task is actioned on.

        Parameters
        ----------
        task
            The ScopedKey for the Task to get actioned AlchemicalNetworks for.
        task_weights
            If ``True``, return a dict with AlchemicalNetwork ScopedKeys as
            keys, the given Task's weights on each AlchemicalNetwork as values.

        Returns
        -------
        networks
            A list of AlchemicalNetwork ScopedKeys which action the given Task.
            If task_weights is ``True``, a dict is returned with
            AlchemicalNetwork ScopedKeys as keys, Task weights as values.

        """
        data = dict(task_weights=task_weights)
        networks = self._post_resource(f"/tasks/{task}/networks/actioned", data)

        if task_weights:
            return {ScopedKey.from_str(n): weight for n, weight in networks.items()}

        return [ScopedKey.from_str(n) for n in networks]

    def action_tasks(
        self,
        tasks: List[ScopedKey],
        network: ScopedKey,
        weight: Optional[Union[float, List[float]]] = None,
    ) -> List[Optional[ScopedKey]]:
        """Action Tasks for execution via the given AlchemicalNetwork's
        TaskHub.

        A Task cannot be actioned:
            - to an AlchemicalNetwork in a different Scope
            - if it has any status other than 'waiting', 'running', or 'error'

        Parameters
        ----------
        tasks
            Task ScopedKeys to action for execution.
        network
            The AlchemicalNetwork ScopedKey to action the Tasks for.
            The Tasks will be added to the network's associated TaskHub.
        weight
            Weight to be applied to the actioned Tasks. Only values between 0
            and 1 are valid weights. Weights can also be provided as a list of
            floats with the same length as `tasks`.

            Setting `weight` to ``None`` will apply the default weight of 0.5
            to newly actioned Tasks, while leaving the weights of any previously
            actioned Tasks unchanged. Setting `weight` to anything other than
            ``None`` will change the weights of previously actioned Tasks
            included in `tasks`.

        Returns
        -------
        List[Optional[ScopedKey]]
            ScopedKeys for Tasks actioned, in the same order as given as
            `tasks` on input. If a Task couldn't be actioned, then ``None``
            will be returned in its place.

        """
        data = dict(tasks=[t.dict() for t in tasks], weight=weight)
        actioned_sks = self._post_resource(f"/networks/{network}/tasks/action", data)

        return [ScopedKey.from_str(i) if i is not None else None for i in actioned_sks]

    def cancel_tasks(
        self, tasks: List[ScopedKey], network: ScopedKey
    ) -> List[ScopedKey]:
        """Cancel Tasks for execution via the given AlchemicalNetwork's
        TaskHub.

        A Task cannot be canceled:
            - if it is not present in the AlchemicalNetwork's TaskHub.

        Parameters
        ----------
        tasks
            Task ScopedKeys to cancel for execution.
        network
            The AlchemicalNetwork ScopedKey to cancel the Tasks for.
            The Tasks will be removed from the network's associated TaskHub.

        Returns
        -------
        List[Optional[ScopedKey]]
            ScopedKeys for Tasks canceled, in the same order as given as
            `tasks` on input. If a Task couldn't be canceled, then ``None`` will
            be returned in its place.

        """
        data = dict(tasks=[t.dict() for t in tasks])
        canceled_sks = self._post_resource(f"/networks/{network}/tasks/cancel", data)

        return [ScopedKey.from_str(i) if i is not None else None for i in canceled_sks]

    def _task_attribute_getter(
        self, tasks: List[ScopedKey], getter_function, batch_size
    ) -> List[Any]:
        tasks = [
            ScopedKey.from_str(task) if isinstance(task, str) else task
            for task in tasks
        ]

        @use_session
        async def async_request(self):
            values = await asyncio.gather(
                *[
                    getter_function(task_batch)
                    for task_batch in self._batched(tasks, batch_size)
                ]
            )

            return list(chain.from_iterable(values))

        coro = async_request(self)

        try:
            return asyncio.run(coro)
        except RuntimeError:
            # we use nest_asyncio to support environments where an event loop
            # is already running, such as in a Jupyter notebook
            import nest_asyncio

            nest_asyncio.apply()
            return asyncio.run(coro)

    def _task_attribute_setter(
        self, tasks: List[ScopedKey], setter_function, setter_args, batch_size
    ) -> List[Optional[ScopedKey]]:
        tasks = [
            ScopedKey.from_str(task) if isinstance(task, str) else task
            for task in tasks
        ]

        @use_session
        async def async_request(self):
            scoped_keys = await asyncio.gather(
                *[
                    setter_function(task_batch, *setter_args)
                    for task_batch in self._batched(tasks, batch_size)
                ]
            )

            return list(chain.from_iterable(scoped_keys))

        coro = async_request(self)

        try:
            return asyncio.run(coro)
        except RuntimeError:
            # we use nest_asyncio to support environments where an event loop
            # is already running, such as in a Jupyter notebook
            import nest_asyncio

            nest_asyncio.apply()
            return asyncio.run(coro)

    async def _set_task_status(
        self, tasks: List[ScopedKey], status: TaskStatusEnum
    ) -> List[Optional[ScopedKey]]:
        """Set the statuses for many Tasks"""
        data = dict(tasks=[t.dict() for t in tasks], status=status.value)
        tasks_updated = await self._post_resource_async(
            f"/bulk/tasks/status/set", data=data
        )
        return [
            ScopedKey.from_str(task_sk) if task_sk is not None else None
            for task_sk in tasks_updated
        ]

    def set_tasks_status(
        self,
        tasks: List[ScopedKey],
        status: Union[TaskStatusEnum, str],
        batch_size: int = 1000,
    ) -> List[Optional[ScopedKey]]:
        """Set the status of one or multiple Tasks.

        Task status can be set to 'waiting' if currently 'error'.
        Status can be set to 'invalid' or 'deleted' from any other status.

        Parameters
        ----------
        tasks
            The Tasks to set the status of.
        status
            The status to set the Tasks to. Can be one of
            'waiting', 'invalid', or 'deleted'.

        Returns
        -------
        updated
            The ScopedKeys of the Tasks that were updated, in the same order
            as given in `tasks`. If a given Task doesn't exist, ``None`` will
            be returned in its place.

        """
        status = TaskStatusEnum(status)

        return self._task_attribute_setter(
            tasks, self._set_task_status, (status,), batch_size
        )

    async def _get_task_status(self, tasks: List[ScopedKey]) -> List[TaskStatusEnum]:
        """Get the statuses for many Tasks"""
        data = dict(tasks=[t.dict() for t in tasks])
        statuses = await self._post_resource_async(f"/bulk/tasks/status/get", data=data)
        return statuses

    def get_tasks_status(
        self, tasks: List[ScopedKey], batch_size: int = 1000
    ) -> List[str]:
        """Get the status of multiple Tasks.

        Parameters
        ----------
        tasks
            The Tasks to get the status of.
        batch_size
            The number of Tasks to include in a single request; use to tune
            method call speed when requesting many statuses at once.

        Returns
        -------
        statuses
            The status of each Task in the same order as given in `tasks`. If a
            given Task doesn't exist, ``None`` will be returned in its place.

        """
        return self._task_attribute_getter(tasks, self._get_task_status, batch_size)

    async def _set_task_priority(
        self, tasks: List[ScopedKey], priority: int
    ) -> List[Optional[ScopedKey]]:
        data = dict(tasks=[t.dict() for t in tasks], priority=priority)
        tasks_updated = await self._post_resource_async(
            f"/bulk/tasks/priority/set", data=data
        )
        return [
            ScopedKey.from_str(task_sk) if task_sk is not None else None
            for task_sk in tasks_updated
        ]

    def set_tasks_priority(
        self,
        tasks: List[ScopedKey],
        priority: int,
        batch_size: int = 1000,
    ) -> List[Optional[ScopedKey]]:
        """Set the priority of multiple Tasks.

        Parameters
        ----------
        tasks
            The Tasks to set the priority of.
        priority
            The priority to set for the Task. This value must be between 1 and
            2**63 - 1, with lower values indicating an increased priority.
        batch_size
            The number of Tasks to include in a single request; use to tune
            method call speed when requesting many priorities at once.

        Returns
        -------
        updated
            The ScopedKeys of the Tasks that were updated, in the same order
            as given in `tasks`. If a given Task doesn't exist, ``None`` will
            be returned in its place.
        """
        return self._task_attribute_setter(
            tasks, self._set_task_priority, (priority,), batch_size
        )

    async def _get_task_priority(self, tasks: List[ScopedKey]) -> List[int]:
        """Get the priority for many Tasks"""
        data = dict(tasks=[t.dict() for t in tasks])
        priorities = await self._post_resource_async(
            f"/bulk/tasks/priority/get", data=data
        )
        return priorities

    def get_tasks_priority(
        self,
        tasks: List[ScopedKey],
        batch_size: int = 1000,
    ) -> List[int]:
        """Get the priority of multiple Tasks.

        Parameters
        ----------
        tasks
            The Tasks to get the priority of.
        batch_size
            The number of Tasks to include in a single request; use to tune
            method call speed when requesting many priorities at once.

        Returns
        -------
        priorities
            The priority of each Task in the same order as given in `tasks`. If a
            given Task doesn't exist, ``None`` will be returned in its place.

        """
        return self._task_attribute_getter(tasks, self._get_task_priority, batch_size)

    ### results

    @alru_cache(maxsize=10000)
    async def _async_get_protocoldagresult(
        self, protocoldagresultref, transformation, route, compress
    ):
        pdr_json = await self._get_resource_async(
            f"/transformations/{transformation}/{route}/{protocoldagresultref}",
            compress=compress,
        )

        pdr = GufeTokenizable.from_dict(
            json.loads(pdr_json[0], cls=JSON_HANDLER.decoder)
        )

        return pdr

    def _get_protocoldagresults(
        self,
        protocoldagresultrefs: List[ScopedKey],
        transformation: ScopedKey,
        ok: bool,
        compress: bool = True,
        visualize: bool = True,
    ):
        if ok:
            route = "results"
        else:
            route = "failures"

        @use_session
        async def async_request(self):
            if visualize:
                from rich.progress import Progress

                with Progress(
                    *self._rich_progress_columns(), transient=False
                ) as progress:
                    task = progress.add_task(
                        f"Retrieving [bold]ProtocolDAGResult[/bold]s",
                        total=len(protocoldagresultrefs),
                    )

                    coros = [
                        self._async_get_protocoldagresult(
                            protocoldagresultref,
                            transformation,
                            route,
                            compress,
                        )
                        for protocoldagresultref in protocoldagresultrefs
                    ]
                    pdrs = []
                    for coro in asyncio.as_completed(coros):
                        pdr = await coro
                        pdrs.append(pdr)
                        progress.update(task, advance=1)
                    progress.refresh()
            else:
                coros = [
                    self._async_get_protocoldagresult(
                        protocoldagresultref,
                        transformation,
                        route,
                        compress,
                    )
                    for protocoldagresultref in protocoldagresultrefs
                ]
                pdrs = await asyncio.gather(*coros)

            return pdrs

        coro = async_request(self)

        try:
            return asyncio.run(coro)
        except RuntimeError:
            # we use nest_asyncio to support environments where an event loop
            # is already running, such as in a Jupyter notebook
            import nest_asyncio

            nest_asyncio.apply()
            return asyncio.run(coro)

    def _get_network_results(
        self,
        network: ScopedKey,
        ok: bool = True,
        return_protocoldagresults: bool = False,
        compress: bool = True,
        visualize: bool = True,
    ) -> Dict[str, Union[Optional[ProtocolResult], List[ProtocolDAGResult]]]:
        import multiprocessing as mp
        from concurrent.futures import ProcessPoolExecutor, as_completed

        ctx = mp.get_context("spawn")

        if ok:
            kwargs = dict(
                return_protocoldagresults=return_protocoldagresults,
                compress=compress,
                visualize=False,
            )
            route = "results"
        else:
            kwargs = dict(compress=compress, visualize=False)
            route = "failures"

        with ProcessPoolExecutor(mp_context=ctx) as executor:
            futures = []
            tf_sks = self.get_network_transformations(network)
            for tf_sk in tf_sks:
                futures.append(
                    executor.submit(
                        _get_transformation_results,
                        self._settings(),
                        tf_sk,
                        ok,
                        kwargs,
                    )
                )

            results = {}
            if visualize:
                from rich.progress import Progress

                with Progress(
                    *self._rich_progress_columns(), transient=False
                ) as progress:
                    task = progress.add_task(
                        f"Retrieving [bold]Transformation[/bold] {route}",
                        total=len(tf_sks),
                    )

                    for future in as_completed(futures):
                        tf_sk, result = future.result()
                        results[tf_sk] = result
                        progress.update(task, advance=1)
                    progress.refresh()
            else:
                for future in as_completed(futures):
                    tf_sk, result = future.result()
                    results[tf_sk] = result

        return results

    def get_network_results(
        self,
        network: ScopedKey,
        return_protocoldagresults: bool = False,
        compress: bool = True,
        visualize: bool = True,
    ) -> Dict[str, Union[Optional[ProtocolResult], List[ProtocolDAGResult]]]:
        """Get a `ProtocolResult` for every `Transformation` in the given
        `AlchemicalNetwork`.

        A dict giving the `ScopedKey` of each `Transformation` in the network
        as keys, `ProtocolResult` as values, is returned. If no
        `ProtocolDAGResult`\s exist for a given `Transformation`, ``None`` is
        given for its value.

        If `return_protocoldagresults` is ``True``, then a list of the
        `ProtocolDAGResult`\s themselves is given as values instead of
        `ProtocolResult`\s.

        Parameters
        ----------
        network
            The `ScopedKey` of the `AlchemicalNetwork` to retrieve results for.
        return_protocoldagresults
            If ``True``, return the raw `ProtocolDAGResult`s instead of returning
            a processed `ProtocolResult`. Only successful `ProtocolDAGResult`\s
            are returned.
        compress
            If ``True``, compress the ProtocolDAGResults server-side before
            shipping them to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicators.

        """
        return self._get_network_results(
            network=network,
            ok=True,
            return_protocoldagresults=return_protocoldagresults,
            compress=compress,
            visualize=visualize,
        )

    def get_network_failures(
        self,
        network: ScopedKey,
        compress: bool = True,
        visualize: bool = True,
    ) -> Dict[str, List[ProtocolDAGResult]]:
        """Get all failed `ProtocolDAGResult`s for every `Transformation` in
        the given `AlchemicalNetwork`.

        A dict giving the `ScopedKey` of each `Transformation` in the network
        as keys, a list of the `ProtocolDAGResult`\s as values, is returned.

        Parameters
        ----------
        network
            The `ScopedKey` of the `AlchemicalNetwork` to retrieve results for.
        compress
            If ``True``, compress the ProtocolDAGResults server-side before
            shipping them to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicators.

        """
        return self._get_network_results(
            network=network, ok=False, compress=compress, visualize=visualize
        )

    def get_transformation_results(
        self,
        transformation: ScopedKey,
        return_protocoldagresults: bool = False,
        compress: bool = True,
        visualize: bool = True,
    ) -> Union[Optional[ProtocolResult], List[ProtocolDAGResult]]:
        """Get a `ProtocolResult` for the given `Transformation`.

        A `ProtocolResult` object corresponds to the `Protocol` used for this
        `Transformation`. This is constructed from the available
        `ProtocolDAGResult`\s for this `Transformation` via
        `Transformation.gather`. If no `ProtocolDAGResult`\s exist for this
        `Transformation`, ``None`` is returned.

        If `return_protocoldagresults` is ``True``, then a list of the
        `ProtocolDAGResult`\s themselves is returned instead.

        Parameters
        ----------
        transformation
            The `ScopedKey` of the `Transformation` to retrieve results for.
        return_protocoldagresults
            If ``True``, return the raw `ProtocolDAGResult`s instead of returning
            a processed `ProtocolResult`. Only successful `ProtocolDAGResult`\s
            are returned.
        compress
            If ``True``, compress the ProtocolDAGResults server-side before
            shipping them to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicators.


        """

        if not return_protocoldagresults:
            # get the transformation if we intend to return a ProtocolResult
            tf: Transformation = self.get_transformation(
                transformation, visualize=visualize
            )

        # get all protocoldagresultrefs for the given transformation
        protocoldagresultrefs = self._get_resource(
            f"/transformations/{transformation}/results",
        )

        pdrs = self._get_protocoldagresults(
            protocoldagresultrefs,
            transformation,
            ok=True,
            compress=compress,
            visualize=visualize,
        )

        if return_protocoldagresults:
            return pdrs
        else:
            if len(pdrs) != 0:
                return tf.gather(pdrs)
            else:
                return None

    def get_transformation_failures(
        self, transformation: ScopedKey, compress: bool = True, visualize: bool = True
    ) -> List[ProtocolDAGResult]:
        """Get failed `ProtocolDAGResult`\s for the given `Transformation`.

        Parameters
        ----------
        transformation
            The `ScopedKey` of the `Transformation` to retrieve failures for.
        compress
            If ``True``, compress the ProtocolDAGResults server-side before
            shipping them to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicators.

        """
        # get all protocoldagresultrefs for the given transformation
        protocoldagresultrefs = self._get_resource(
            f"/transformations/{transformation}/failures",
        )

        pdrs = self._get_protocoldagresults(
            protocoldagresultrefs,
            transformation,
            ok=False,
            compress=compress,
            visualize=visualize,
        )

        return pdrs

    def get_task_results(
        self, task: ScopedKey, compress: bool = True, visualize: bool = True
    ) -> List[ProtocolDAGResult]:
        """Get successful `ProtocolDAGResult`s for the given `Task`.

        Parameters
        ----------
        task
            The `ScopedKey` of the `Task` to retrieve results for.
        compress
            If ``True``, compress the ProtocolDAGResults server-side before
            shipping them to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicators.

        """
        # first, get the transformation; also confirms it exists
        transformation: ScopedKey = self.get_task_transformation(task)

        # get all protocoldagresultrefs for the given transformation
        protocoldagresultrefs = self._get_resource(
            f"/tasks/{task}/results",
        )

        pdrs = self._get_protocoldagresults(
            protocoldagresultrefs,
            transformation,
            ok=True,
            compress=compress,
            visualize=visualize,
        )

        return pdrs

    def get_task_failures(
        self, task: ScopedKey, compress: bool = True, visualize: bool = True
    ) -> List[ProtocolDAGResult]:
        """Get failed `ProtocolDAGResult`s for the given `Task`.

        Parameters
        ----------
        task
            The `ScopedKey` of the `Task` to retrieve failures for.
        compress
            If ``True``, compress the ProtocolDAGResults server-side before
            shipping them to the client. This can reduce retrieval time depending
            on the bandwidth of your connection to the API service. Set to
            ``False`` to retrieve without compressing. This is a performance
            optimization; it has no bearing on the result of this method call.
        visualize
            If ``True``, show retrieval progress indicators.

        """
        # first, get the transformation; also confirms it exists
        transformation: ScopedKey = self.get_task_transformation(task)

        # get all protocoldagresultrefs for the given transformation
        protocoldagresultrefs = self._get_resource(
            f"/tasks/{task}/failures",
        )

        pdrs = self._get_protocoldagresults(
            protocoldagresultrefs,
            transformation,
            ok=False,
            compress=compress,
            visualize=visualize,
        )

        return pdrs



"""
:mod:`alchemiscale.validators` --- validation guardrails for user input
=======================================================================

"""

from gufe import AlchemicalNetwork, Transformation


def validate_network_nonself(network: AlchemicalNetwork):
    """Check that the given AlchemicalNetwork features no Transformations with
    the same ChemicalSystem for its two states.

    A ``ValueError`` is raised if a `Transformation` is detected.

    """
    for transformation in network.edges:
        if transformation.stateA == transformation.stateB:
            raise ValueError(
                f"`Transformation` '{transformation.key}' uses the same `ChemicalSystem` '{transformation.stateA.key}' for both states; "
                "this is currently not supported in `alchemiscale`"
            )



"""
:mod:`alchemiscale.cli` --- command line interface
==================================================

"""

import click
import yaml
import json
import signal
from typing import Type

from .security.models import (
    CredentialedEntity,
    CredentialedUserIdentity,
    CredentialedComputeIdentity,
)


def envvar_dictify(ctx, param, value):
    """Callback to return a dict of param's envvar to value.

    This ensures that the envvar name only has to be entered as a string
    once within the click system. It requires that the parameter this
    callback is attached to defines its envvar.
    """
    return {param.envvar: value}


# use these extra kwargs with any option from a settings parameter.
SETTINGS_OPTION_KWARGS = {
    "show_envvar": True,
    "callback": envvar_dictify,
}


def get_settings_from_options(kwargs, settings_cls):
    """Create a settings object from a dict.

    This first strips all items with value None (which will be defaults) so
    that they don't override settings defaults.
    """
    update = {k: v for k, v in kwargs.items() if v is not None}
    return settings_cls(**update)


def api_starting_params(envvar_host, envvar_port, envvar_loglevel):
    def inner(func):
        workers = click.option(
            "--workers", type=int, help="number of workers", default=1
        )
        host = click.option(
            "--host",
            type=str,
            help="IP address of host",
            envvar=envvar_host,
            **SETTINGS_OPTION_KWARGS,
        )
        port = click.option(
            "--port",
            type=int,
            help="port",
            envvar=envvar_port,
            **SETTINGS_OPTION_KWARGS,
        )
        loglevel = click.option(
            "--loglevel",
            type=str,
            default="info",
            help="logging level",
            envvar=envvar_loglevel,
            **SETTINGS_OPTION_KWARGS,
        )
        config_file = click.option(
            "--config-file",
            "-c",
            type=click.File(),
            help="YAML-based configuration file giving additional settings for gunicorn",
        )
        config_json = click.option(
            "--config-json",
            "-j",
            type=str,
            help="inline JSON giving additional settings for gunicorn; these take precedence over options given with `--config-file`",
        )
        return workers(host(port(loglevel(config_file(config_json(func))))))

    return inner


def db_params(func):
    url = click.option(
        "--url",
        help="database URI",
        type=str,
        envvar="NEO4J_URL",
        required=True,
        **SETTINGS_OPTION_KWARGS,
    )
    user = click.option(
        "--user",
        help="database user name",
        type=str,
        envvar="NEO4J_USER",
        required=True,
        **SETTINGS_OPTION_KWARGS,
    )
    password = click.option(
        "--password",
        help="database password",
        type=str,
        envvar="NEO4J_PASS",
        required=True,
        **SETTINGS_OPTION_KWARGS,
    )
    dbname = click.option(
        "--dbname",
        type=str,
        help="custom database name, default 'neo4j'",
        envvar="NEO4J_DBNAME",
        **SETTINGS_OPTION_KWARGS,
    )
    return url(user(password(dbname(func))))


def generate_secret_key(ctx, param, value):
    from alchemiscale.security.auth import generate_secret_key

    return {param.envvar: generate_secret_key()}


def jwt_params(func):
    secret = click.option(
        "--jwt-secret",
        type=str,
        help="JSON web token secret",
        envvar="JWT_SECRET_KEY",
        callback=generate_secret_key,
        show_envvar=True,
    )
    expire_seconds = click.option(
        "--jwt-expire-seconds",
        type=int,
        default=1800,
        envvar="JWT_EXPIRE_SECONDS",
        **SETTINGS_OPTION_KWARGS,
    )
    algo = click.option(
        "--jwt-algorithm",
        type=str,
        default="HS256",
        envvar="JWT_ALGORITHM",
        **SETTINGS_OPTION_KWARGS,
    )
    return secret(expire_seconds(algo(func)))


def s3os_params(func):
    access_key_id = click.option(
        "--access-key-id",
        type=str,
        default=None,
        envvar="AWS_ACCESS_KEY_ID",
        **SETTINGS_OPTION_KWARGS,
    )
    secret_access_key = click.option(
        "--secret-access-key",
        type=str,
        default=None,
        envvar="AWS_SECRET_ACCESS_KEY",
        **SETTINGS_OPTION_KWARGS,
    )
    session_token = click.option(
        "--session-token",
        type=str,
        default=None,
        envvar="AWS_SESSION_TOKEN",
        **SETTINGS_OPTION_KWARGS,
    )
    s3_bucket = click.option(
        "--s3-bucket",
        type=str,
        envvar="AWS_S3_BUCKET",
        required=True,
        **SETTINGS_OPTION_KWARGS,
    )
    s3_prefix = click.option(
        "--s3-prefix",
        type=str,
        envvar="AWS_S3_PREFIX",
        required=True,
        **SETTINGS_OPTION_KWARGS,
    )
    default_region = click.option(
        "--default-region",
        type=str,
        envvar="AWS_DEFAULT_REGION",
        required=True,
        **SETTINGS_OPTION_KWARGS,
    )
    return access_key_id(
        secret_access_key(session_token(s3_bucket(s3_prefix(default_region(func)))))
    )


def start_api(api_app, workers, host, port, options):
    from .cli_utils import ApiApplication

    gunicorn_app = ApiApplication(
        api_app, workers, bind=f"{host}:{port}", options=options
    )
    gunicorn_app.run()


@click.group()
def cli():
    ...


# reusable parameters to ensure consistent naming and help strings


@cli.command(
    name="api",
    help="Start the user-facing API service",
)
@api_starting_params(
    "ALCHEMISCALE_API_HOST", "ALCHEMISCALE_API_PORT", "ALCHEMISCALE_API_LOGLEVEL"
)
@db_params
@s3os_params
@jwt_params
def api(
    workers, host, port, loglevel, config_file, config_json,  # API
    url, user, password, dbname,  # DB
    jwt_secret, jwt_expire_seconds, jwt_algorithm,  # JWT
    access_key_id, secret_access_key, session_token, s3_bucket, s3_prefix, default_region  # AWS
):  # fmt: skip
    from alchemiscale.interface.api import app
    from .settings import APISettings, get_base_api_settings
    from .security.auth import generate_secret_key

    # CONSIDER GENERATING A JWT_SECRET_KEY if none provided with
    # key = generate_secret_key()
    # CONVENIENT FOR THE SINGLE-SERVER CASE HERE
    # HOW-TO: modify the callback in jwt_secret (defined in jwt_params) to
    # do this. See comment there. Use that instead of the callback in
    # SETTINGS_OPTION_KWARGS

    options = {}
    if config_file is not None:
        options.update(yaml.safe_load(config_file))

    if config_json is not None:
        options.update(json.loads(config_json))

    def get_settings_override():
        # inject settings from CLI arguments
        api_dict = host | port | loglevel
        jwt_dict = jwt_secret | jwt_expire_seconds | jwt_algorithm
        db_dict = url | user | password | dbname
        s3_dict = (
            access_key_id
            | secret_access_key
            | session_token
            | s3_bucket
            | s3_prefix
            | default_region
        )
        return get_settings_from_options(
            api_dict | jwt_dict | db_dict | s3_dict, APISettings
        )

    app.dependency_overrides[get_base_api_settings] = get_settings_override

    start_api(
        app,
        workers,
        host["ALCHEMISCALE_API_HOST"],
        port["ALCHEMISCALE_API_PORT"],
        options=options,
    )


@cli.group(help="Subcommands for compute services")
def compute():
    ...


@compute.command(help="Start the compute API service.")
@api_starting_params(
    "ALCHEMISCALE_COMPUTE_API_HOST",
    "ALCHEMISCALE_COMPUTE_API_PORT",
    "ALCHEMISCALE_COMPUTE_API_LOGLEVEL",
)
@click.option(
    "--registration-expire-seconds",
    type=int,
    default=1800,
    help="number of seconds since last heartbeat at which to expire a compute service registration",
    envvar="ALCHEMISCALE_COMPUTE_API_REGISTRATION_EXPIRE_SECONDS",
    **SETTINGS_OPTION_KWARGS,
)
@db_params
@s3os_params
@jwt_params
def api(
    workers, host, port, loglevel, config_file, config_json, registration_expire_seconds, # API
    url, user, password, dbname,  # DB
    jwt_secret, jwt_expire_seconds, jwt_algorithm,  #JWT
    access_key_id, secret_access_key, session_token, s3_bucket, s3_prefix, default_region  # AWS
):  # fmt: skip
    from alchemiscale.compute.api import app
    from .settings import ComputeAPISettings, get_base_api_settings
    from .security.auth import generate_secret_key

    # CONSIDER GENERATING A JWT_SECRET_KEY if none provided with
    # key = generate_secret_key()
    # CONVENIENT FOR THE SINGLE-SERVER CASE HERE

    options = {}
    if config_file is not None:
        options.update(yaml.safe_load(config_file))

    if config_json is not None:
        options.update(json.loads(config_json))

    def get_settings_override():
        # inject settings from CLI arguments
        api_dict = host | port | loglevel | registration_expire_seconds
        jwt_dict = jwt_secret | jwt_expire_seconds | jwt_algorithm
        db_dict = url | user | password | dbname
        s3_dict = (
            access_key_id
            | secret_access_key
            | session_token
            | s3_bucket
            | s3_prefix
            | default_region
        )
        return get_settings_from_options(
            api_dict | jwt_dict | db_dict | s3_dict, ComputeAPISettings
        )

    app.dependency_overrides[get_base_api_settings] = get_settings_override

    start_api(
        app,
        workers,
        host["ALCHEMISCALE_COMPUTE_API_HOST"],
        port["ALCHEMISCALE_COMPUTE_API_PORT"],
        options=options,
    )


@compute.command(help="Start the synchronous compute service.")
@click.option(
    "--config-file",
    "-c",
    type=click.File(),
    help="YAML-based configuration file giving the settings for this service",
    required=True,
)
def synchronous(config_file):
    from alchemiscale.models import Scope
    from alchemiscale.compute.service import SynchronousComputeService

    params = yaml.safe_load(config_file)

    params_init = params.get("init", {})
    params_start = params.get("start", {})

    if "scopes" in params_init:
        params_init["scopes"] = [
            Scope.from_str(scope) for scope in params_init["scopes"]
        ]

    service = SynchronousComputeService(**params_init)

    # add signal handling
    for signame in {"SIGHUP", "SIGINT", "SIGTERM"}:

        def stop(*args, **kwargs):
            service.stop()
            raise KeyboardInterrupt()

        signal.signal(getattr(signal, signame), stop)

    try:
        service.start(**params_start)
    except KeyboardInterrupt:
        pass


@cli.group(help="Subcommands for the database")
def database():
    ...


@database.command()
@db_params
def init(url, user, password, dbname):
    """Initialize the Neo4j database.

    Note that options here can be set by environment variables, as shown on
    each option.
    """
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname
    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)

    n4js = get_n4js(settings)
    n4js.initialize()


@database.command()
@db_params
def check(url, user, password, dbname):
    """Check consistency of database.

    Note that options here can be set by environment variables, as shown on
    each option.
    """
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    db_dict = url | user | password | dbname

    settings = get_settings_from_options(db_dict, Neo4jStoreSettings)

    n4js = get_n4js(settings)
    if n4js.check() is None:
        print("No inconsistencies found in database.")


@database.command()
@db_params
def reset(url, user, password, dbname):
    """Remove all data from database; undo `init`.

    Note that options here can be set by environment variables, as shown on
    each option.
    """
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname
    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)

    n4js = get_n4js(settings)
    n4js.reset()


def _identity_type_string_to_cls(identity_type: str) -> Type[CredentialedEntity]:
    if identity_type == "user":
        identity_type_cls = CredentialedUserIdentity
    elif identity_type == "compute":
        identity_type_cls = CredentialedComputeIdentity
    else:
        raise RuntimeError(f"Unknown identity type {identity_type}")

    return identity_type_cls


def identity_type(func):
    identity_type = click.option(
        "--identity-type",
        "-t",
        default="user",
        help="User type",
        type=click.Choice(["user", "compute"], case_sensitive=False),
    )
    return identity_type(func)


def identifier(func):
    identifier = click.option(
        "--identifier", "-i", help="identifier", required=True, type=str
    )
    return identifier((func))


def key(func):
    key = click.option("--key", "-k", help="key", required=True, type=str)
    return key(func)


def scope(func):
    scope = click.option(
        "--scope", "-s", help="scope", required=True, type=str, multiple=True
    )
    return scope(func)


@cli.group(help="Subcommands for managing identities")
def identity():
    ...


@identity.command()
@db_params
@identity_type
@identifier
@key
def add(url, user, password, dbname, identity_type, identifier, key):
    """Add a credentialed identity to the database."""
    from .security.auth import hash_key
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname

    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)
    n4js = get_n4js(settings)

    identity_type_cls = _identity_type_string_to_cls(identity_type)
    identity_model = identity_type_cls(hashed_key=hash_key(key), identifier=identifier)
    n4js.create_credentialed_entity(identity_model)


@identity.command()
@db_params
@identity_type
def list(url, user, password, dbname, identity_type):
    """List all credentialed entities of the given type."""
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname

    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)
    n4js = get_n4js(settings)

    identity_type_cls = _identity_type_string_to_cls(identity_type)
    click.echo(n4js.list_credentialed_entities(identity_type_cls))


@identity.command()
@db_params
@identity_type
@identifier
def remove(url, user, password, dbname, identity_type, identifier):
    """Remove a credentialed identity from the database."""
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname

    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)
    n4js = get_n4js(settings)

    identity_type_cls = _identity_type_string_to_cls(identity_type)
    n4js.remove_credentialed_identity(identifier, identity_type_cls)


@identity.command()
@db_params
@identity_type
@identifier
@scope
def add_scope(url, user, password, dbname, identity_type, identifier, scope):
    """Add a scope for the given identity."""
    from .models import Scope
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname

    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)
    n4js = get_n4js(settings)

    identity_type_cls = _identity_type_string_to_cls(identity_type)
    for s in scope:
        s_ = Scope.from_str(s)
        n4js.add_scope(identifier, identity_type_cls, s_)


@identity.command()
@db_params
@identity_type
@identifier
def list_scope(url, user, password, dbname, identity_type, identifier):
    """List all scopes for the given identity."""
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname

    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)
    n4js = get_n4js(settings)

    identity_type_cls = _identity_type_string_to_cls(identity_type)
    scopes = n4js.list_scopes(identifier, identity_type_cls)

    click.echo([str(scope) for scope in scopes])


@identity.command()
@db_params
@identity_type
@identifier
@scope
def remove_scope(url, user, password, dbname, identity_type, identifier, scope):
    """Remove a scope for the given identity(s)."""
    from .models import Scope
    from .storage.statestore import get_n4js
    from .settings import Neo4jStoreSettings

    cli_values = url | user | password | dbname

    settings = get_settings_from_options(cli_values, Neo4jStoreSettings)
    n4js = get_n4js(settings)

    identity_type_cls = _identity_type_string_to_cls(identity_type)

    for s in scope:
        s_ = Scope.from_str(s)
        n4js.remove_scope(identifier, identity_type_cls, s_)



"""
:mod:`alchemiscale.storage.objectstore` --- object store interface
==================================================================

"""

import os
import io
import json
from datetime import datetime
from typing import Union, Optional
from boto3.session import Session
from functools import lru_cache

from gufe.protocols import ProtocolDAGResult
from gufe.tokenization import JSON_HANDLER, GufeTokenizable

from ..models import ScopedKey, Scope
from .models import ProtocolDAGResultRef
from ..settings import S3ObjectStoreSettings, get_s3objectstore_settings


@lru_cache()
def get_s3os(settings: S3ObjectStoreSettings, endpoint_url=None) -> "S3ObjectStore":
    """Convenience function for getting an S3ObjectStore directly from settings."""

    # create a boto3 Session and parameterize with keys
    session = Session(
        aws_access_key_id=settings.AWS_ACCESS_KEY_ID,
        aws_secret_access_key=settings.AWS_SECRET_ACCESS_KEY,
        aws_session_token=settings.AWS_SESSION_TOKEN,
        region_name=settings.AWS_DEFAULT_REGION,
    )

    return S3ObjectStore(
        session=session,
        bucket=settings.AWS_S3_BUCKET,
        prefix=settings.AWS_S3_PREFIX,
        endpoint_url=endpoint_url,
    )


class S3ObjectStoreError(Exception):
    ...


class S3ObjectStore:
    """Object storage for use with AWS S3."""

    def __init__(
        self, session: "boto3.Session", bucket: str, prefix: str, endpoint_url=None
    ):
        """ """
        self.session = session
        self.resource = self.session.resource("s3", endpoint_url=endpoint_url)

        self.bucket = bucket
        self.prefix = prefix

    def initialize(self):
        """Initialize object store.

        Creates bucket if it does not exist.

        """
        bucket = self.resource.Bucket(self.bucket)
        bucket.create()
        bucket.wait_until_exists()

    def check(self):
        """Check consistency of object store."""
        raise NotImplementedError

    def _store_check(self):
        """Check that the ObjectStore is in a state that can be used by the API."""
        try:
            # read check
            self.resource.meta.client.list_buckets()

            # write check
            self._store_bytes("_check_test", b"test_check")
            self._delete("_check_test")
        except:
            return False
        return True

    def reset(self):
        """Remove all data from object store.

        Deletes all objects, including the bucket itself.

        """
        bucket = self.resource.Bucket(self.bucket)

        # delete all objects, then the bucket
        bucket.objects.delete()
        bucket.delete()
        bucket.wait_until_not_exists()

    def iter_contents(self, prefix=""):
        """Iterate over the labels in this storage.

        Parameters
        ----------
        prefix : str
            Only iterate over paths that start with the given prefix.

        Returns
        -------
        Iterator[str] :
            Contents of this storage, which may include items without
            metadata.
        """

        filter_prefix = os.path.join(self.prefix, prefix)

        return self.resource.Bucket(self.bucket).objects.filter(Prefix=filter_prefix)

    def _store_bytes(self, location, byte_data):
        """
        For implementers: This should be blocking, even if the storage
        backend allows asynchronous storage.
        """
        key = os.path.join(self.prefix, location)

        response = self.resource.Object(self.bucket, key).put(Body=byte_data)

        if not response["ResponseMetadata"]["HTTPStatusCode"] == 200:
            raise S3ObjectStoreError(f"Could not store given object at key {key}")

        return response

    def _get_bytes(self, location):
        key = os.path.join(self.prefix, location)

        b = self.resource.Bucket(self.bucket)

        return self.resource.Object(self.bucket, key).get()["Body"].read()

    def _store_path(self, location, path):
        """
        For implementers: This should be blocking, even if the storage
        backend allows asynchronous storage.
        """
        """
        For implementers: This should be blocking, even if the storage
        backend allows asynchronous storage.
        """
        key = os.path.join(self.prefix, location)

        with open(path, "rb") as f:
            self.resource.Bucket(self.bucket).upload_fileobj(f, key)

        b = self.resource.Bucket(self.bucket)

    def _exists(self, location) -> bool:
        from botocore.exceptions import ClientError

        key = os.path.join(self.prefix, location)

        # we do a metadata load as our existence check
        # appears to be most recommended approach
        try:
            self.resource.Object(self.bucket, key).load()
            return True
        except ClientError:
            return False

    def _delete(self, location):
        key = os.path.join(self.prefix, location)

        if self._exists(location):
            self.resource.Object(self.bucket, key).delete()
        else:
            raise S3ObjectStoreError(
                f"Unable to delete '{str(key)}': Object does not exist"
            )

    def _get_filename(self, location):
        key = os.path.join(self.prefix, location)

        object = self.bucket.Object(key)

        url = object.meta.client.generate_presigned_url(
            "get_object",
            ExpiresIn=0,
            Params={"Bucket": self.bucket.name, "Key": object.key},
        )

        # drop query params from url
        url = url.split("?")[0]

        return url

    def push_protocoldagresult(
        self,
        protocoldagresult: ProtocolDAGResult,
        transformation: ScopedKey,
        creator: Optional[str] = None,
    ) -> ProtocolDAGResultRef:
        """Push given `ProtocolDAGResult` to this `ObjectStore`.

        Parameters
        ----------
        protocoldagresult
            ProtocolDAGResult to store.
        transformation
            The ScopedKey of the Transformation this ProtocolDAGResult
            corresponds to.

        Returns
        -------
        ProtocolDAGResultRef
            Reference to the serialized `ProtocolDAGResult` in the object store.

        """
        ok = protocoldagresult.ok()
        route = "results" if ok else "failures"

        # build `location` based on gufe key
        location = os.path.join(
            "protocoldagresult",
            *transformation.scope.to_tuple(),
            transformation.gufe_key,
            route,
            protocoldagresult.key,
            "obj.json",
        )

        # TODO: add support for compute client-side compressed protocoldagresults
        pdr_jb = json.dumps(
            protocoldagresult.to_dict(), cls=JSON_HANDLER.encoder
        ).encode("utf-8")
        response = self._store_bytes(location, pdr_jb)

        return ProtocolDAGResultRef(
            location=location,
            obj_key=protocoldagresult.key,
            scope=transformation.scope,
            ok=ok,
            datetime_created=datetime.utcnow(),
            creator=creator,
        )

    def pull_protocoldagresult(
        self,
        protocoldagresult: Optional[ScopedKey] = None,
        transformation: Optional[ScopedKey] = None,
        location: Optional[str] = None,
        return_as="gufe",
        ok=True,
    ) -> Union[ProtocolDAGResult, dict, str]:
        """Pull the `ProtocolDAGResult` corresponding to the given `ProtocolDAGResultRef`.

        Parameters
        ----------
        protocoldagresult
            ScopedKey for ProtocolDAGResult in the object store.
            Must be provided if `location` is ``None``.
        transformation
            The ScopedKey of the Transformation this ProtocolDAGResult
            corresponds to.
            Must be provided if `location` is ``None``.
        location
            The full path in the object store to the ProtocolDAGResult. If
            provided, this will be used to retrieve it.
        return_as : ['gufe', 'dict', 'json']
            Form in which to return result; this is provided to avoid
            unnecessary deserializations where desired.

        Returns
        -------
        ProtocolDAGResult
            The ProtocolDAGResult corresponding to the given `ProtocolDAGResultRef`.

        """
        route = "results" if ok else "failures"

        # build `location` based on provided ScopedKey if not provided
        if location is None:
            if None in (transformation, protocoldagresult):
                raise ValueError(
                    "`transformation` and `protocoldagresult` must both be given if `location` is ``None``"
                )
            if transformation.scope != protocoldagresult.scope:
                raise ValueError(
                    f"transformation scope '{transformation.scope}' differs from protocoldagresult scope '{protocoldagresult.scope}'"
                )

            location = os.path.join(
                "protocoldagresult",
                *protocoldagresult.scope.to_tuple(),
                transformation.gufe_key,
                route,
                protocoldagresult.gufe_key,
                "obj.json",
            )

        ## TODO: want organization alongside `obj.json` of `ProtocolUnit` gufe_keys
        ## for any file objects stored in the same space

        pdr_j = self._get_bytes(location).decode("utf-8")

        # TODO: add support for interface client-side decompression
        if return_as == "gufe":
            pdr = GufeTokenizable.from_dict(json.loads(pdr_j, cls=JSON_HANDLER.decoder))
        elif return_as == "dict":
            pdr = json.loads(pdr_j, cls=JSON_HANDLER.decoder)
        elif return_as == "json":
            pdr = pdr_j

        return pdr



import gunicorn.app.base


class ApiApplication(gunicorn.app.base.BaseApplication):
    def __init__(self, app, workers, bind, options=None):
        self.app = app
        self.workers = workers
        self.bind = bind
        self.options = options or {}
        super().__init__()

    @classmethod
    def from_parameters(cls, app, workers, host, port, options=None):
        return cls(app, workers, bind=f"{host}:{port}", options=options)

    def load(self):
        return self.app

    def load_config(self):
        self.cfg.set("workers", self.workers)
        self.cfg.set("bind", self.bind)
        self.cfg.set("worker_class", "uvicorn.workers.UvicornWorker")

        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }

        for key, value in config.items():
            self.cfg.set(key.lower(), value)



"""
:mod:`alchemiscale.interface.api` --- user-facing API components
================================================================

"""

from typing import Any, Dict, List, Optional, Union
import os
import json
from collections import Counter

from fastapi import FastAPI, APIRouter, Body, Depends, HTTPException, status
from fastapi.middleware.gzip import GZipMiddleware
from gufe import AlchemicalNetwork, ChemicalSystem, Transformation
from gufe.protocols import ProtocolDAGResult
from gufe.tokenization import GufeTokenizable, JSON_HANDLER

from ..base.api import (
    GufeJSONResponse,
    QueryGUFEHandler,
    scope_params,
    get_token_data_depends,
    get_n4js_depends,
    get_s3os_depends,
    base_router,
    get_cred_entity,
    validate_scopes,
    validate_scopes_query,
    _check_store_connectivity,
    gufe_to_json,
    GzipRoute,
)
from ..settings import get_api_settings
from ..settings import get_base_api_settings, get_api_settings
from ..storage.statestore import Neo4jStore
from ..storage.objectstore import S3ObjectStore
from ..storage.models import ProtocolDAGResultRef, TaskStatusEnum
from ..models import Scope, ScopedKey
from ..security.auth import get_token_data, oauth2_scheme
from ..security.models import Token, TokenData, CredentialedUserIdentity
from ..keyedchain import KeyedChain


app = FastAPI(title="AlchemiscaleAPI")
app.dependency_overrides[get_base_api_settings] = get_api_settings
app.include_router(base_router)
app.add_middleware(GZipMiddleware, minimum_size=1000, compresslevel=5)


def get_cred_user():
    return CredentialedUserIdentity


app.dependency_overrides[get_cred_entity] = get_cred_user

router = APIRouter(
    dependencies=[Depends(get_token_data_depends)],
)
router.route_class = GzipRoute


@app.get("/ping")
def ping():
    return {"api": "AlchemiscaleAPI"}


@router.get("/info")
def info():
    return {"message": "nothing yet"}


@router.get("/check")
def check(
    n4js: Neo4jStore = Depends(get_n4js_depends),
    s3os: S3ObjectStore = Depends(get_s3os_depends),
):
    # check connectivity of storage components
    # if no exception raised, all good
    _check_store_connectivity(n4js, s3os)


@router.get("/identities/{identity_identifier}/scopes")
def list_scopes(
    *,
    identity_identifier,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[str]:
    scopes = n4js.list_scopes(identity_identifier, CredentialedUserIdentity)
    return [str(scope) for scope in scopes]


### inputs


@router.get("/exists/{scoped_key}")
def check_existence(
    scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(scoped_key)
    validate_scopes(sk.scope, token)

    return n4js.check_existence(scoped_key=sk)


@router.post("/networks", response_model=ScopedKey)
def create_network(
    *,
    network: List = Body(...),
    scope: Scope,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    validate_scopes(scope, token)

    an = KeyedChain(network).to_gufe()

    try:
        an_sk = n4js.create_network(network=an, scope=scope)
    except ValueError as e:
        raise HTTPException(
            status_code=status.HTTP_422_UNPROCESSABLE_ENTITY,
            detail=e.args[0],
        )

    # create taskhub for this network
    n4js.create_taskhub(an_sk)

    return an_sk


@router.get("/networks")
def query_networks(
    *,
    name: str = None,
    scope: Scope = Depends(scope_params),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    # Intersect query scopes with accessible scopes in the token
    query_scopes = validate_scopes_query(scope, token)

    # query each scope
    # loop might be removable in the future with a Union like operator on scopes
    results = []
    for single_query_scope in query_scopes:
        results.extend(
            n4js.query_networks(
                name=name,
                scope=single_query_scope,
            )
        )

    return [str(sk) for sk in results]


@router.get("/transformations")
def query_transformations(
    *,
    name: str = None,
    scope: Scope = Depends(scope_params),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    # Intersect query scopes with accessible scopes in the token
    query_scopes = validate_scopes_query(scope, token)

    # query each scope
    # loop might be removable in the future with a Union like operator on scopes
    results = []
    for single_query_scope in query_scopes:
        results.extend(n4js.query_transformations(name=name, scope=single_query_scope))

    return [str(sk) for sk in results]


@router.get("/chemicalsystems")
def query_chemicalsystems(
    *,
    name: str = None,
    scope: Scope = Depends(scope_params),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    # Intersect query scopes with accessible scopes in the token
    query_scopes = validate_scopes_query(scope, token)

    # query each scope
    # loop might be removable in the future with a Union like operator on scopes
    results = []
    for single_query_scope in query_scopes:
        results.extend(n4js.query_chemicalsystems(name=name, scope=single_query_scope))

    return [str(sk) for sk in results]


@router.get("/networks/{network_scoped_key}/weight")
def get_network_weight(
    network_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> float:
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)
    return n4js.get_taskhub_weight(sk)


@router.get("/networks/{network_scoped_key}/transformations")
def get_network_transformations(
    network_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_network_transformations(network=sk)]


@router.get("/transformations/{transformation_scoped_key}/networks")
def get_transformation_networks(
    transformation_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_transformation_networks(transformation=sk)]


@router.get("/networks/{network_scoped_key}/chemicalsystems")
def get_network_chemicalsystems(
    network_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_network_chemicalsystems(network=sk)]


@router.get("/chemicalsystems/{chemicalsystem_scoped_key}/networks")
def get_chemicalsystem_networks(
    chemicalsystem_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(chemicalsystem_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_chemicalsystem_networks(chemicalsystem=sk)]


@router.get("/transformations/{transformation_scoped_key}/chemicalsystems")
def get_transformation_chemicalsystems(
    transformation_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    return [
        str(sk) for sk in n4js.get_transformation_chemicalsystems(transformation=sk)
    ]


@router.get("/chemicalsystems/{chemicalsystem_scoped_key}/transformations")
def get_chemicalsystem_transformations(
    chemicalsystem_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(chemicalsystem_scoped_key)
    validate_scopes(sk.scope, token)

    return [
        str(sk) for sk in n4js.get_chemicalsystem_transformations(chemicalsystem=sk)
    ]


@router.get("/networks/{network_scoped_key}")
def get_network(
    network_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    network = n4js.get_gufe(scoped_key=sk)
    return GufeJSONResponse(network)


@router.get("/transformations/{transformation_scoped_key}")
def get_transformation(
    transformation_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    transformation = n4js.get_gufe(scoped_key=sk)
    return GufeJSONResponse(transformation)


@router.get("/chemicalsystems/{chemicalsystem_scoped_key}")
def get_chemicalsystem(
    chemicalsystem_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(chemicalsystem_scoped_key)
    validate_scopes(sk.scope, token)

    chemicalsystem = n4js.get_gufe(scoped_key=sk)
    return GufeJSONResponse(chemicalsystem)


### compute


@router.post("/networks/{scoped_key}/strategy")
def set_strategy(scoped_key: str, *, strategy: Dict = Body(...), scope: Scope):
    ...


@router.post("/transformations/{transformation_scoped_key}/tasks")
def create_tasks(
    transformation_scoped_key,
    *,
    extends: Optional[ScopedKey] = None,
    count: int = Body(...),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[str]:
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    task_sks = []
    for i in range(count):
        task_sks.append(
            n4js.create_task(transformation=sk, extends=extends, creator=token.entity)
        )

    return [str(sk) for sk in task_sks]


@router.get("/tasks")
def query_tasks(
    *,
    status: str = None,
    scope: Scope = Depends(scope_params),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    # Intersect query scopes with accessible scopes in the token
    query_scopes = validate_scopes_query(scope, token)

    # query each scope
    # loop might be removable in the future with a Union like operator on scopes
    results = []
    for single_query_scope in query_scopes:
        results.extend(n4js.query_tasks(status=status, scope=single_query_scope))

    return [str(sk) for sk in results]


@router.get("/networks/{network_scoped_key}/tasks")
def get_network_tasks(
    network_scoped_key,
    *,
    status: str = None,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    # Get scope from scoped key provided by user, uniquely identifying the network
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    if status is not None:
        status = TaskStatusEnum(status)

    return [str(sk) for sk in n4js.get_network_tasks(network=sk, status=status)]


@router.get("/tasks/{task_scoped_key}/networks")
def get_task_networks(
    task_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_task_networks(task=sk)]


@router.get("/transformations/{transformation_scoped_key}/tasks")
def get_transformation_tasks(
    transformation_scoped_key,
    *,
    extends: str = None,
    return_as: str = "list",
    status: str = None,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    if status is not None:
        status = TaskStatusEnum(status)

    task_sks = n4js.get_transformation_tasks(
        sk, extends=extends, return_as=return_as, status=status
    )

    if return_as == "list":
        return [str(sk) for sk in task_sks]
    elif return_as == "graph":
        return {
            str(sk): str(extends) if extends is not None else None
            for sk, extends in task_sks.items()
        }
    else:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"`return_as` takes 'list' or 'graph', not '{return_as}'",
        )


@router.get("/scopes/{scope}/status")
def get_scope_status(
    scope,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    scope = Scope.from_str(scope)
    scope_space = validate_scopes_query(scope, token)

    status_counts = Counter()
    for single_scope in scope_space:
        status_counts.update(n4js.get_scope_status(single_scope))

    return dict(status_counts)


@router.get("/networks/{network_scoped_key}/status")
def get_network_status(
    network_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    status_counts = n4js.get_network_status(network_scoped_key)

    return status_counts


@router.get("/transformations/{transformation_scoped_key}/status")
def get_transformation_status(
    transformation_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    status_counts = n4js.get_transformation_status(transformation_scoped_key)

    return status_counts


@router.post("/networks/{network_scoped_key}/tasks/actioned")
def get_network_actioned_tasks(
    network_scoped_key,
    *,
    task_weights: bool = Body(embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> Union[Dict[str, float], List[str]]:
    network_sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(network_sk.scope, token)

    taskhub_sk = n4js.get_taskhub(network_sk)
    task_sks = n4js.get_taskhub_actioned_tasks(taskhub_sk)

    if task_weights:
        return {str(task_sk): weight for task_sk, weight in task_sks.items()}

    return [str(task_sk) for task_sk in task_sks]


@router.post("/tasks/{task_scoped_key}/networks/actioned")
def get_task_actioned_networks(
    task_scoped_key,
    *,
    task_weights: bool = Body(embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> Union[Dict[str, float], List[str]]:
    task_sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(task_sk.scope, token)

    network_sks = n4js.get_task_actioned_networks(task_sk)

    if task_weights:
        return {str(network_sk): weight for network_sk, weight in network_sks.items()}

    return [str(network_sk) for network_sk in network_sks]


@router.post("/networks/{network_scoped_key}/tasks/action")
def action_tasks(
    network_scoped_key,
    *,
    tasks: List[ScopedKey] = Body(embed=True),
    weight: Optional[Union[float, List[float]]] = Body(None, embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[Union[str, None]]:
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    taskhub_sk = n4js.get_taskhub(sk)
    actioned_sks = n4js.action_tasks(tasks, taskhub_sk)

    try:
        if isinstance(weight, float):
            n4js.set_task_weights(tasks, taskhub_sk, weight)
        elif isinstance(weight, list):
            if len(weight) != len(tasks):
                detail = "weight (when in a list) must have the same length as tasks"
                raise HTTPException(
                    status_code=status.HTTP_400_BAD_REQUEST,
                    detail=detail,
                )

            n4js.set_task_weights(
                {task: weight for task, weight in zip(tasks, weight)}, taskhub_sk, None
            )
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))

    return [str(sk) if sk is not None else None for sk in actioned_sks]


@router.post("/networks/{network_scoped_key}/weight")
def set_network_weight(
    network_scoped_key,
    *,
    weight: float = Body(),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> None:
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    try:
        n4js.set_taskhub_weight(sk, weight)
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))


@router.post("/networks/{network_scoped_key}/tasks/cancel")
def cancel_tasks(
    network_scoped_key,
    *,
    tasks: List[ScopedKey] = Body(embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[Union[str, None]]:
    sk = ScopedKey.from_str(network_scoped_key)
    validate_scopes(sk.scope, token)

    taskhub_sk = n4js.get_taskhub(sk)
    canceled_sks = n4js.cancel_tasks(tasks, taskhub_sk)

    return [str(sk) if sk is not None else None for sk in canceled_sks]


@router.post("/bulk/tasks/priority/get")
def tasks_priority_get(
    *,
    tasks: List[ScopedKey] = Body(embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[int]:
    valid_tasks = []
    for task_sk in tasks:
        try:
            validate_scopes(task_sk.scope, token)
            valid_tasks.append(task_sk)
        except HTTPException:
            valid_tasks.append(None)

    priorities = n4js.get_task_priority(valid_tasks)

    return priorities


@router.post("/bulk/tasks/priority/set")
def tasks_priority_set(
    *,
    tasks: List[ScopedKey] = Body(embed=True),
    priority: int = Body(embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[Union[str, None]]:
    valid_tasks = []
    for task_sk in tasks:
        try:
            validate_scopes(task_sk.scope, token)
            valid_tasks.append(task_sk)
        except HTTPException:
            valid_tasks.append(None)

    try:
        tasks_updated = n4js.set_task_priority(valid_tasks, priority)
    except ValueError as e:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))

    return [str(t) if t is not None else None for t in tasks_updated]


@router.post("/bulk/tasks/status/get")
def tasks_status_get(
    *,
    tasks: List[ScopedKey] = Body(embed=True),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[Union[str, None]]:
    valid_tasks = []
    for task_sk in tasks:
        try:
            validate_scopes(task_sk.scope, token)
            valid_tasks.append(task_sk)
        except HTTPException:
            valid_tasks.append(None)

    statuses = n4js.get_task_status(valid_tasks)

    return [status.value if status is not None else None for status in statuses]


@router.post("/bulk/tasks/status/set")
def tasks_status_set(
    *,
    tasks: List[ScopedKey] = Body(),
    status: str = Body(),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[Union[str, None]]:
    status = TaskStatusEnum(status)
    if status not in (
        TaskStatusEnum.waiting,
        TaskStatusEnum.invalid,
        TaskStatusEnum.deleted,
    ):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Cannot set status to '{status}', must be one of 'waiting', 'invalid', 'deleted'",
        )

    valid_tasks = []
    for task_sk in tasks:
        try:
            validate_scopes(task_sk.scope, token)
            valid_tasks.append(task_sk)
        except HTTPException:
            valid_tasks.append(None)

    tasks_updated = n4js.set_task_status(valid_tasks, status)

    return [str(t) if t is not None else None for t in tasks_updated]


@router.post("/tasks/{task_scoped_key}/status")
def set_task_status(
    task_scoped_key,
    status: str = Body(),
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    status = TaskStatusEnum(status)
    if status not in (
        TaskStatusEnum.waiting,
        TaskStatusEnum.invalid,
        TaskStatusEnum.deleted,
    ):
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Cannot set status to '{status}', must be one of 'waiting', 'invalid', 'deleted'",
        )
    task_sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(task_sk.scope, token)
    tasks_statused = n4js.set_task_status([task_sk], status)
    return [str(t) if t is not None else None for t in tasks_statused][0]


@router.get("/tasks/{task_scoped_key}/status")
def get_task_status(
    task_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    task_sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(task_sk.scope, token)

    status = n4js.get_task_status([task_sk])

    return status[0].value


@router.get("/tasks/{task_scoped_key}/transformation")
def get_task_transformation(
    task_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(sk.scope, token)

    transformation: ScopedKey

    transformation, protocoldagresultref = n4js.get_task_transformation(
        task=task_scoped_key,
        return_gufe=False,
    )

    return str(transformation)


### results


@router.get("/transformations/{transformation_scoped_key}/results")
def get_transformation_results(
    transformation_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_transformation_results(sk)]


@router.get("/transformations/{transformation_scoped_key}/failures")
def get_transformation_failures(
    transformation_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(transformation_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_transformation_failures(sk)]


@router.get(
    "/transformations/{transformation_scoped_key}/{route}/{protocoldagresultref_scoped_key}"
)
def get_protocoldagresult(
    protocoldagresultref_scoped_key,
    route,
    transformation_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    s3os: S3ObjectStore = Depends(get_s3os_depends),
    token: TokenData = Depends(get_token_data_depends),
) -> List[str]:
    if route == "results":
        ok = True
    elif route == "failures":
        ok = False
    else:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"`route` takes 'results' or 'failures', not '{route}'",
        )

    sk = ScopedKey.from_str(protocoldagresultref_scoped_key)
    transformation_sk = ScopedKey.from_str(transformation_scoped_key)

    validate_scopes(sk.scope, token)
    validate_scopes(transformation_sk.scope, token)

    protocoldagresultref = n4js.get_gufe(scoped_key=sk)
    pdr_sk = ScopedKey(gufe_key=protocoldagresultref.obj_key, **sk.scope.dict())

    # we leave each ProtocolDAGResult in string form to avoid
    # deserializing/reserializing here; just passing through to client
    try:
        pdr: str = s3os.pull_protocoldagresult(
            pdr_sk, transformation_sk, return_as="json", ok=ok
        )
    except:
        # if we fail to get the object with the above, fall back to
        # location-based retrieval
        pdr: str = s3os.pull_protocoldagresult(
            location=protocoldagresultref.location,
            return_as="json",
            ok=ok,
        )

    return [pdr]


@router.get("/tasks/{task_scoped_key}/results")
def get_task_results(
    task_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_task_results(sk)]


@router.get("/tasks/{task_scoped_key}/failures")
def get_task_failures(
    task_scoped_key,
    *,
    n4js: Neo4jStore = Depends(get_n4js_depends),
    token: TokenData = Depends(get_token_data_depends),
):
    sk = ScopedKey.from_str(task_scoped_key)
    validate_scopes(sk.scope, token)

    return [str(sk) for sk in n4js.get_task_failures(sk)]


### add router

app.include_router(router)



"""
:mod:`alchemiscale.security.models` --- data models for security components
===========================================================================

"""

from datetime import datetime, timedelta
from typing import List, Union, Optional

from pydantic import BaseModel, validator

from ..models import Scope


class Token(BaseModel):
    access_token: str
    token_type: str


class TokenData(BaseModel):
    entity: Optional[str] = None
    scopes: List[str] = None


class CredentialedEntity(BaseModel):
    hashed_key: str
    expires: Optional[datetime] = None


class ScopedIdentity(BaseModel):
    identifier: str
    disabled: bool = False
    scopes: List[str] = []

    @validator("scopes", pre=True, each_item=True)
    def cast_scopes_to_str(cls, scope):
        """Ensure that each scope object is correctly cast to its str representation"""
        if isinstance(scope, Scope):
            scope = str(scope)
        elif isinstance(scope, str):
            try:
                Scope.from_str(scope)
            except:
                raise ValueError(f"Invalid scope `{scope}` set for `{cls}`")
        else:
            raise ValueError(f"Invalid scope `{scope}` set for `{cls}`")

        return scope


class UserIdentity(ScopedIdentity):
    email: Optional[str] = None
    full_name: Optional[str] = None


class CredentialedUserIdentity(UserIdentity, CredentialedEntity):
    ...


class ComputeIdentity(ScopedIdentity):
    email: Optional[str] = None


class CredentialedComputeIdentity(ComputeIdentity, CredentialedEntity):
    ...



from gufe.tokenization import GufeTokenizable, key_decode_dependencies
import networkx as nx
from alchemiscale.utils import gufe_to_digraph

from typing import List, Tuple, Dict, Generator


class KeyedChain(object):
    """Keyed chain representation of a GufeTokenizable.

    The keyed chain representation of a GufeTokenizable provides a
    topologically sorted list of gufe keys and GufeTokenizable keyed dicts
    that can be used to fully recreate a GufeTokenizable without the need for a
    populated TOKENIZATION_REGISTRY.

    The class wraps around a list of tuples containing the gufe key and the
    keyed dict form of the GufeTokenizable.

    """

    def __init__(self, keyed_chain):
        self._keyed_chain = keyed_chain

    @classmethod
    def from_gufe(cls, gufe_object: GufeTokenizable) -> super:
        """Initialize a KeyedChain from a GufeTokenizable."""
        return cls(cls.gufe_to_keyed_chain_rep(gufe_object))

    def to_gufe(self) -> GufeTokenizable:
        """Initialize a GufeTokenizable."""
        gts = {}
        for gufe_key, keyed_dict in self:
            gt = key_decode_dependencies(keyed_dict, registry=gts)
            gts[gufe_key] = gt
        return gt

    @staticmethod
    def gufe_to_keyed_chain_rep(
        gufe_object: GufeTokenizable,
    ) -> List[Tuple[str, Dict]]:
        """Create the keyed chain represenation of a GufeTokenizable.

        This represents the GufeTokenizable as a list of two-element tuples
        containing, as their first and second elements, the gufe key and keyed
        dict form of the GufeTokenizable, respectively, and provides the
        underlying structure used in the KeyedChain class.

        Parameters
        ----------
        gufe_object
            The GufeTokenizable for which the KeyedChain is generated.

        Returns
        -------
        key_and_keyed_dicts
            The keyed chain represenation of a GufeTokenizable.

        """
        key_and_keyed_dicts = [
            (str(gt.key), gt.to_keyed_dict())
            for gt in nx.topological_sort(gufe_to_digraph(gufe_object))
        ][::-1]
        return key_and_keyed_dicts

    def gufe_keys(self) -> Generator[str, None, None]:
        """Create a generator that iterates over the gufe keys in the KeyedChain."""
        for key, _ in self:
            yield key

    def keyed_dicts(self) -> Generator[Dict, None, None]:
        """Create a generator that iterates over the keyed dicts in the KeyedChain."""
        for _, _dict in self:
            yield _dict

    def __len__(self):
        return len(self._keyed_chain)

    def __iter__(self):
        return self._keyed_chain.__iter__()

    def __getitem__(self, index):
        return self._keyed_chain[index]
